from qtpy import QtGui
from qtpy import QtCore
from qtpy import QtWidgets

from typstwriter import util

from typstwriter import logging
from typstwriter import configuration
from typstwriter import globalstate

logger = logging.getLogger(__name__)
config = configuration.Config
state = globalstate.State


class Actions(QtCore.QObject):
    """Stores the commonly used actions."""

    def __init__(self, parent):
        """Init."""
        super().__init__(parent)

        self.new_File = QtWidgets.QAction(self)
        self.new_File.setIcon(QtGui.QIcon.fromTheme(QtGui.QIcon.DocumentNew, QtGui.QIcon(util.icon_path("newFile.svg"))))
        self.new_File.setShortcut(QtGui.QKeySequence.New)
        self.new_File.setText("New File")

        self.open_File = QtWidgets.QAction(self)
        self.open_File.setIcon(QtGui.QIcon.fromTheme(QtGui.QIcon.DocumentOpen, QtGui.QIcon(util.icon_path("openFile.svg"))))
        self.open_File.setShortcut(QtGui.QKeySequence.Open)
        self.open_File.setText("Open File")

        self.open_recent_File = QtWidgets.QAction(self)
        self.open_recent_File.setIcon(
            QtGui.QIcon.fromTheme(QtGui.QIcon.DocumentOpenRecent, QtGui.QIcon(util.icon_path("recentFile.svg")))
        )
        self.open_recent_File.setShortcut(QtGui.QKeySequence(QtCore.Qt.CTRL | QtCore.Qt.SHIFT | QtCore.Qt.Key_O))
        self.open_recent_File.setText("Open Recent File")

        self.load_last_Session = QtWidgets.QAction(self)
        self.load_last_Session.setIcon(
            QtGui.QIcon.fromTheme("folder-open-recent", QtGui.QIcon(util.icon_path("lastSession.svg")))
        )
        self.load_last_Session.setText("Load last Session")

        self.save = QtWidgets.QAction(self)
        self.save.setIcon(QtGui.QIcon.fromTheme(QtGui.QIcon.DocumentSave, QtGui.QIcon(util.icon_path("save.svg"))))
        self.save.setShortcut(QtGui.QKeySequence.Save)
        self.save.setText("Save")

        self.save_as = QtWidgets.QAction(self)
        self.save_as.setIcon(QtGui.QIcon.fromTheme(QtGui.QIcon.DocumentSaveAs, QtGui.QIcon(util.icon_path("save_as.svg"))))
        self.save_as.setShortcut(QtGui.QKeySequence(QtCore.Qt.CTRL | QtCore.Qt.SHIFT | QtCore.Qt.Key_S))
        self.save_as.setText("Save As")

        self.close = QtWidgets.QAction(self)
        self.close.setIcon(QtGui.QIcon.fromTheme("document-close-symbolic", QtGui.QIcon(util.icon_path("closeFile.svg"))))
        self.close.setShortcut(QtGui.QKeySequence.Close)
        self.close.setText("Close")

        self.quit = QtWidgets.QAction(self)
        self.quit.setIcon(QtGui.QIcon.fromTheme(QtGui.QIcon.ApplicationExit, QtGui.QIcon(util.icon_path("quit.svg"))))
        self.quit.setShortcut(QtGui.QKeySequence.Quit)
        self.quit.setText("Quit")

        self.cut = QtWidgets.QAction(self)
        self.cut.setIcon(QtGui.QIcon.fromTheme(QtGui.QIcon.EditCut, QtGui.QIcon(util.icon_path("cut.svg"))))
        self.cut.setShortcut(QtGui.QKeySequence.Cut)
        self.cut.setText("Cut")

        self.copy = QtWidgets.QAction(self)
        self.copy.setIcon(QtGui.QIcon.fromTheme(QtGui.QIcon.EditCopy, QtGui.QIcon(util.icon_path("copy.svg"))))
        self.copy.setShortcut(QtGui.QKeySequence.Copy)
        self.copy.setShortcutContext(QtCore.Qt.WidgetWithChildrenShortcut)
        self.copy.setText("Copy")

        self.paste = QtWidgets.QAction(self)
        self.paste.setIcon(QtGui.QIcon.fromTheme(QtGui.QIcon.EditPaste, QtGui.QIcon(util.icon_path("paste.svg"))))
        self.paste.setShortcut(QtGui.QKeySequence.Paste)
        self.paste.setShortcutContext(QtCore.Qt.WidgetWithChildrenShortcut)
        self.paste.setText("Paste")

        self.search = QtWidgets.QAction(self)
        self.search.setIcon(QtGui.QIcon.fromTheme(QtGui.QIcon.EditFind, QtGui.QIcon(util.icon_path("search.svg"))))
        self.search.setShortcut(QtGui.QKeySequence.Find)
        self.search.setText("Search")

        self.font_size_up = QtWidgets.QAction(self)
        self.font_size_up.setIcon(QtGui.QIcon.fromTheme(QtGui.QIcon.ZoomIn, QtGui.QIcon(util.icon_path("plus.svg"))))
        self.font_size_up.setText("Increase Font Size")

        self.font_size_dn = QtWidgets.QAction(self)
        self.font_size_dn.setIcon(QtGui.QIcon.fromTheme(QtGui.QIcon.ZoomOut, QtGui.QIcon(util.icon_path("minus.svg"))))
        self.font_size_dn.setText("Decrease Font Size")

        self.font_size_reset = QtWidgets.QAction(self)
        self.font_size_reset.setIcon(QtGui.QIcon.fromTheme(QtGui.QIcon.ZoomFitBest))
        self.font_size_reset.setText("Reset Font Size")

        self.layout_typewriter = QtWidgets.QAction(self)
        self.layout_typewriter.setIcon(QtGui.QIcon.fromTheme("view-split-top-bottom-symbolic"))
        self.layout_typewriter.setText("Layout: Typewriter")
        self.layout_typewriter.setCheckable(True)

        self.layout_editorL = QtWidgets.QAction(self)
        self.layout_editorL.setIcon(QtGui.QIcon.fromTheme("view-split-left-right-symbolic"))
        self.layout_editorL.setText("Layout: Editor Left")
        self.layout_editorL.setCheckable(True)

        self.layout_editorR = QtWidgets.QAction(self)
        self.layout_editorR.setIcon(QtGui.QIcon.fromTheme("view-split-left-right-symbolic"))
        self.layout_editorR.setText("Layout: Editor Right")
        self.layout_editorR.setCheckable(True)

        self.layout = QtWidgets.QActionGroup(self)
        self.layout.addAction(self.layout_typewriter)
        self.layout.addAction(self.layout_editorL)
        self.layout.addAction(self.layout_editorR)

        self.show_fs_explorer = QtWidgets.QAction(self)
        self.show_fs_explorer.setText("Show FS Explorer")
        self.show_fs_explorer.setCheckable(True)

        self.show_compiler_options = QtWidgets.QAction(self)
        self.show_compiler_options.setText("Show Compiler Options")
        self.show_compiler_options.setCheckable(True)

        self.show_compiler_output = QtWidgets.QAction(self)
        self.show_compiler_output.setText("Show Compiler Output")
        self.show_compiler_output.setCheckable(True)

        self.open_config = QtWidgets.QAction(self)
        self.open_config.setIcon(QtGui.QIcon.fromTheme("configure-symbolic"))
        self.open_config.setText("Open config file")

        self.run = util.TogglingAction(self)
        self.run.setIcon(
            QtGui.QIcon.fromTheme(QtGui.QIcon.MediaPlaybackStart, QtGui.QIcon(util.icon_path("start.svg"))),
            state=QtGui.QIcon.State.Off,
        )
        self.run.setIcon(
            QtGui.QIcon.fromTheme(QtGui.QIcon.MediaPlaybackStop, QtGui.QIcon(util.icon_path("stop.svg"))),
            state=QtGui.QIcon.State.On,
        )
        self.run.setText("Start", state=QtGui.QIcon.State.Off)
        self.run.setText("Stop", state=QtGui.QIcon.State.On)
        self.run.setShortcut("Ctrl+R")



import enum

compiler_mode = enum.Enum("compiler_mode", ["on_demand", "live"])
search_mode = enum.Enum("search_mode", ["case_insensitive", "case_sensitive", "whole_words", "regex"])
search_direction = enum.Enum("search_direction", ["next", "previous"])



# from typstwriter import compiler
# from qtpy import QtCore
# import time
# import pytest


# def mock_process_start


# class TestCompilerConnector_FS_onDemand:
#
#     # @pytest.fixture(scope="class")
#     # def patch(self, monkeypatch):
#     #     """Patch the QtCore.QProcess.start() function to do nothing."""
#     #     monkeypatch.setattr(QtCore.QProcess, "start", lambda: None)
#
#     def test_successful_compilation(self, tmp_path, qtbot):
#         """Test a successful compilation"""
#         fin = str(tmp_path / "fin")
#         fout = str(tmp_path / "fout")
#
#         c = compiler.CompilerConnector_FS_onDemand(fin=fin, fout=fout)
#
#         with qtbot.waitSignal(c.started):
#             c.start()

#
# class TestCompilerConnector_FS_live:
#     pass
#
#
# class TestWrappedCompilerConnector:
#     pass




from qtpy import QtCore
from qtpy import QtGui

import pytest

from typstwriter import editor
from typstwriter import enums

search_data = [
    (
        "Fischers Fritz fischt frische Fische,\nfrische Fische fischt Fischers Fritz",
        "Fisch",
        enums.search_mode.case_insensitive,
        [(0, 5), (15, 20), (30, 35), (46, 51), (53, 58), (60, 65)],
        "_",
        "_ers Fritz _t frische _e,\nfrische _e _t _ers Fritz",
    ),
    (
        "Fischers Fritz fischt frische Fische,\nfrische Fische fischt Fischers Fritz",
        "Fisch",
        enums.search_mode.case_sensitive,
        [(0, 5), (30, 35), (46, 51), (60, 65)],
        "_",
        "_ers Fritz fischt frische _e,\nfrische _e fischt _ers Fritz",
    ),
    (
        "Fischers Fritz fischt frische Fische,\nfrische Fische fischt Fischers Fritz",
        "Fische",
        enums.search_mode.whole_words,
        [(30, 36), (46, 52)],
        "_",
        "Fischers Fritz fischt frische _,\nfrische _ fischt Fischers Fritz",
    ),
    (
        "Fischers Fritz fischt frische Fische,\nfrische Fische fischt Fischers Fritz",
        r"\bFischers \w+\b",
        enums.search_mode.regex,
        [(0, 14), (60, 74)],
        "_",
        "_ fischt frische Fische,\nfrische Fische fischt _",
    ),
]

comment_text = [
    ("Just some text.", "//Just some text.", "Just some text.", (0, 0)),
    ("Just\nsome\ntext.", "//Just\n//some\ntext.", "Just\nsome\ntext.", (0, 1)),
    ("Just\n//some\ntext.", "//Just\n////some\n//text.", "Just\n//some\ntext.", (0, 2)),
    ("Just\n//some\ntext.", "Just\nsome\ntext.", "Just\n//some\ntext.", (1, 1)),
    ("Just\n  //some\ntext.", "Just\n  some\ntext.", "Just\n//  some\ntext.", (1, 1)),
    ("Just\nsome//more\ntext.", "Just\n//some//more\ntext.", "Just\nsome//more\ntext.", (1, 1)),
]


class TestWelcomePage:
    """Test editor.WelcomePage."""

    def test_update_recent_files(self, qtbot):
        """Test updating the recent files."""
        recent_files = ["A.typ", "B.typ", "C.txt"]

        welcome_page = editor.WelcomePage()
        welcome_page.update_recent_files(recent_files)

        assert welcome_page.recentFilesModel.recent_files == recent_files
        for i, f in enumerate(recent_files):
            assert welcome_page.recentFilesModel.itemData(welcome_page.recentFilesModel.index(i))[0] == f

    def test_open_file(self, qtbot, tmp_path):
        """Test doubleclick in recent files."""
        path = tmp_path / "file.typ"
        path.touch()

        welcome_page = editor.WelcomePage()
        welcome_page.update_recent_files([str(path)])
        qtbot.addWidget(welcome_page)

        with qtbot.waitSignal(welcome_page.open_file) as blocker:
            welcome_page.doubleclicked(welcome_page.recentFilesModel.index(0))
        assert blocker.args[0] == str(path)

    def test_tryclose(self):
        """Test trying to close the Page."""
        welcome_page = editor.WelcomePage()
        assert welcome_page.tryclose() is True


class TestCodeEdit:
    """Test editor.CodeEdit."""

    # TODO: Test syntax highlighting
    # TODO: Test line numbers

    def test_line_highlighting_on(self, qtbot):
        """Test line highlighting."""
        code_edit = editor.CodeEdit(highlight_synatx=False, show_line_numbers=False, highlight_line=True)
        code_edit.insertPlainText("Just\nsome\nexample\ntext.")

        # Selecting a single line should highlight
        code_edit.moveCursor(QtGui.QTextCursor.Start)
        code_edit.moveCursor(QtGui.QTextCursor.NextBlock, QtGui.QTextCursor.MoveMode.MoveAnchor)
        assert code_edit.extraSelections()[0].cursor == code_edit.textCursor()

        # Selecting multiple lines should not highlight
        code_edit.moveCursor(QtGui.QTextCursor.Start)
        code_edit.moveCursor(QtGui.QTextCursor.NextBlock, QtGui.QTextCursor.MoveMode.KeepAnchor)
        assert not code_edit.extraSelections()

    def test_line_highlighting_off(self, qtbot):
        """Test line highlighting."""
        code_edit = editor.CodeEdit(highlight_synatx=False, show_line_numbers=False, highlight_line=False)
        code_edit.insertPlainText("Just\nsome\nexample\ntext.")
        code_edit.moveCursor(QtGui.QTextCursor.Start)
        code_edit.moveCursor(QtGui.QTextCursor.NextBlock)
        assert not code_edit.extraSelections()

    def test_key_press_tab(self, qtbot):
        """Test key press interception when pressing tab."""
        code_edit = editor.CodeEdit(highlight_synatx=False, show_line_numbers=False, use_spaces=True)
        qtbot.keyPress(code_edit, QtCore.Qt.Key_Tab)
        assert code_edit.toPlainText() == "    "

    def test_key_press_linebreak(self, qtbot):
        """Test key press interception when pressing shift enter."""
        code_edit = editor.CodeEdit(highlight_synatx=False, show_line_numbers=False, use_spaces=True)
        qtbot.keyPress(code_edit, QtCore.Qt.Key_Return, QtCore.Qt.ShiftModifier)
        assert code_edit.toPlainText() == "\n"

    @pytest.mark.parametrize(("text", "query", "mode", "_spans", "replace_text", "result"), search_data)
    def test_replace_all_matches(self, query, mode, text, replace_text, result, _spans):  # noqa PT019
        """Test replace_all_matches() for all four query modes."""
        code_edit = editor.CodeEdit(highlight_synatx=False, show_line_numbers=False, use_spaces=True)
        code_edit.insertPlainText(text)
        code_edit.replace_all_matches(query, mode, replace_text)
        assert code_edit.toPlainText() == result

    @pytest.mark.parametrize(("text", "query", "mode", "spans", "_replace_text", "_result"), search_data)
    def test_highlight_all_matches(self, query, mode, text, spans, _replace_text, _result):  # noqa PT019
        """Test highlight_all_matches() for all four query modes."""
        code_edit = editor.CodeEdit(highlight_synatx=False, show_line_numbers=False, use_spaces=True)
        code_edit.insertPlainText(text)
        code_edit.highlight_all_matches(query, mode)
        for selection, span in zip(code_edit.search_highlights, spans, strict=True):
            assert selection.cursor.anchor() == span[0]
            assert selection.cursor.position() == span[1]

    @pytest.mark.parametrize(("text", "toggle", "retoggle", "selection"), comment_text)
    def test_highlight_all_matches(self, text, toggle, retoggle, selection):  # noqa PT019
        """Test toggle_comment() for various selections."""
        start, stop = selection
        code_edit = editor.CodeEdit(highlight_synatx=False, show_line_numbers=False, use_spaces=True)
        code_edit.insertPlainText(text)
        cursor = code_edit.textCursor()
        cursor.setPosition(0)
        cursor.movePosition(QtGui.QTextCursor.NextBlock, QtGui.QTextCursor.MoveAnchor, start)
        cursor.movePosition(QtGui.QTextCursor.NextBlock, QtGui.QTextCursor.KeepAnchor, stop - start)
        code_edit.setTextCursor(cursor)

        code_edit.toggle_comment()
        assert code_edit.toPlainText() == toggle

        code_edit.toggle_comment()
        assert code_edit.toPlainText() == retoggle



import pytest

from typstwriter import globalstate


class TestSingleState:
    """Test globalstate.SingleState."""

    def test_Value_assignment(self, qtbot):
        """Make sure Value can be assigned."""
        state = globalstate.SingleState("Test")
        assert state.Value == "Test"

    def test_Signal_emitted(self, qtbot):
        """Make sure Value assignment triggers Signal."""
        state = globalstate.SingleState("")
        with qtbot.waitSignal(state.Signal) as blocker:
            state.Value = "Test"
        assert blocker.args == ["Test"]

    def test_AttributeError(self, qtbot):
        """Make sure other fields trigger AttributeError."""
        state = globalstate.SingleState("")
        with pytest.raises(AttributeError, match="The only valid attributes are 'Value' and 'Signal'."):
            state.Test = "Test"


class TestGloalState:
    """Test globalstate.GlobalState."""

    def test_members(self, qtbot):
        """Test if all required members are present."""
        state = globalstate.GlobalState()

        assert hasattr(state, "working_directory")
        assert hasattr(state, "compiler_running")
        assert hasattr(state, "main_file")
        assert hasattr(state, "compiler_mode")



from qtpy import QtGui
from qtpy import QtCore
from qtpy import QtWidgets

import os

from typstwriter import enums
from typstwriter import util

from typstwriter import logging
from typstwriter import configuration
from typstwriter import globalstate

logger = logging.getLogger(__name__)
config = configuration.Config
state = globalstate.State


class CompilerOptions(QtWidgets.QWidget):
    """Displays the compiler options."""

    def __init__(self):
        """Populate the widget."""
        QtWidgets.QWidget.__init__(self)

        self.FrameLayout = QtWidgets.QVBoxLayout(self)
        self.FrameLayout.setContentsMargins(4, 4, 4, 4)
        self.FrameLayout.setSpacing(2)

        self.HostWidget = QtWidgets.QFrame(self)
        self.HostWidget.setFrameStyle(QtWidgets.QFrame.Shape.StyledPanel)
        palette = self.HostWidget.palette()
        palette.setColor(QtGui.QPalette.ColorRole.Window, QtGui.QColor("white"))
        self.HostWidget.setAutoFillBackground(True)
        self.HostWidget.setPalette(palette)
        self.FrameLayout.addWidget(self.HostWidget)

        self.Layout = QtWidgets.QGridLayout(self.HostWidget)
        self.Layout.setContentsMargins(4, 4, 4, 4)
        self.Layout.setSpacing(2)
        self.Layout.setAlignment(QtGui.Qt.AlignTop)

        # Mode
        self.label_mode = QtWidgets.QLabel("Mode")
        self.combo_box_mode = QtWidgets.QComboBox()
        self.combo_box_mode.addItem("On Demand", enums.compiler_mode.on_demand)
        self.combo_box_mode.addItem("Live", enums.compiler_mode.live)
        self.combo_box_mode.currentIndexChanged.connect(self.mode_changed)
        self.combo_box_mode.setCurrentIndex(self.combo_box_mode.findData(state.compiler_mode.Value))

        # Main file
        self.label_main = QtWidgets.QLabel("Main file")
        self.line_edit_main = QtWidgets.QLineEdit(self)
        self.line_edit_main.setText("")
        self.line_edit_main.editingFinished.connect(self.main_path_edited)
        self.folderAction = QtWidgets.QAction(
            QtGui.QIcon.fromTheme(QtGui.QIcon.FolderOpen, QtGui.QIcon(util.icon_path("folder.svg"))), "open"
        )
        self.folderAction.triggered.connect(self.open_file_dialog)
        self.line_edit_main.addAction(self.folderAction, QtWidgets.QLineEdit.LeadingPosition)

        # Insert everything into layout
        self.Layout.addWidget(self.label_mode, 0, 0)
        self.Layout.addWidget(self.combo_box_mode, 0, 1)
        self.Layout.addWidget(self.label_main, 1, 0)
        self.Layout.addWidget(self.line_edit_main, 1, 1)

    @QtCore.Slot()
    def main_path_edited(self):
        """Handle LineEdits editingFinished signal. Change main file path if it is valid and resets otherwise."""
        path = os.path.normpath(os.path.expanduser(self.line_edit_main.text()))
        if os.path.isfile(path):
            state.main_file.Value = path
        else:
            self.line_edit_main.setText(state.main_file.Value)
            logger.info("Attempted to set main file but {!r} is not a valid path.", path)

    @QtCore.Slot()
    def open_file_dialog(self):
        """Open a dialog to select the main file."""
        filters = "Typst Files (*.typ);;Any File (*)"
        path, cd = QtWidgets.QFileDialog.getOpenFileName(self, "Open File", state.working_directory.Value, filters)

        if os.path.isfile(path):
            self.line_edit_main.setText(path)
            state.main_file.Value = path

    @QtCore.Slot(str)
    def main_changed(self, path):
        """Update display of main file."""
        self.line_edit_main.setText(path)

    @QtCore.Slot(int)
    def mode_changed(self, index):
        """Update the compiler mode."""
        mode = self.combo_box_mode.itemData(index)
        state.compiler_mode.Value = mode


class CompilerOutput(QtWidgets.QWidget):
    """Displays the compiler output."""

    def __init__(self):
        """Populate the widget up."""
        QtWidgets.QWidget.__init__(self)

        self.OutputDisplay = QtWidgets.QTextEdit()
        self.OutputDisplay.setReadOnly(True)

        self.ClearDisplay = QtWidgets.QPushButton()
        self.ClearDisplay.setText("Clear Output")
        self.ClearDisplay.pressed.connect(self.OutputDisplay.clear)

        self.Layout = QtWidgets.QVBoxLayout(self)
        self.Layout.setContentsMargins(4, 4, 4, 4)
        self.Layout.setSpacing(2)
        self.Layout.addWidget(self.OutputDisplay)
        self.Layout.addWidget(self.ClearDisplay)

    def insert_block(self, text=None):
        """Insert a new block of text."""
        self.insert_text(new_block=True, text=text)

    def append_to_block(self, text):
        """Append to the current block of text."""
        self.insert_text(new_block=False, text=text)

    def insert_text(self, new_block=False, text=None):
        """Insert text."""
        bottom_scrolled = self.OutputDisplay.verticalScrollBar().value() == self.OutputDisplay.verticalScrollBar().maximum()

        if new_block and not self.OutputDisplay.document().isEmpty():
            self.OutputDisplay.insertHtml("<p/><hr/><p/>")

        if text:
            self.OutputDisplay.insertPlainText(text)

        if bottom_scrolled:
            self.OutputDisplay.verticalScrollBar().setValue(self.OutputDisplay.verticalScrollBar().maximum())



import os
import configparser

import platformdirs

from typstwriter import logging

logger = logging.getLogger(__name__)


default_config_path = os.path.join(platformdirs.user_config_dir("typstwriter", "typstwriter"), "typstwriter.ini")
default_recent_files_path = os.path.join(platformdirs.user_data_dir("typstwriter", "typstwriter"), "recentFiles.txt")
default_session_path = os.path.join(platformdirs.user_data_dir("typstwriter", "typstwriter"), "Session.txt")

config_paths = ["/etc/typstwriter/typstwriter.ini",
                "/usr/local/etc/typstwriter/typstwriter.ini",
                default_config_path,
                os.path.expanduser("~/.typstwriter.ini"),
                "./typstwriter.ini"]  # fmt: skip

default_config = {"General": {"working_directory": "~/",
                              "resume_last_session": False},
                  "Compiler": {"name": "typst",
                               "mode": "on_demand"},
                  "Editor": {"font_size": 10,
                             "save_at_run": False,
                             "highlighter_style": "sas",  # Can be any style from https://pygments.org/styles/
                             "highlight_syntax": True,
                             "show_line_numbers": True,
                             "highlight_line": True,
                             "use_spaces": True},
                  "Layout": {"default_layout": "typewriter",
                             "show_fs_explorer": True,
                             "show_compiler_options": True,
                             "show_compiler_output": True},
                  "Internals": {"recent_files_path": default_recent_files_path,
                                "recent_files_length": 16,
                                "session_path": default_session_path}}  # fmt: skip


class ConfigManager:
    """Handle configuration."""

    def __init__(self, paths=None):
        """Set config to default and attempt to read config files, if given."""
        self.config = configparser.ConfigParser(empty_lines_in_values=False)

        # load default config
        self.config.read_dict(default_config)

        self.readpaths = paths
        self.writepath = default_config_path

        # read config
        if paths:
            usedfile = self.config.read(self.readpaths)

            # Check if read was successful
            if usedfile:
                self.writepath = usedfile[-1]
            else:
                logger.warning("No valid config file found in {!r}", paths)

    def get(self, section, key, typ="str"):
        """Get config value."""
        match typ:
            case "int" | "integer":
                return self.config.getint(section, key)
            case "float":
                return self.config.getfloat(section, key)
            case "bool" | "boolean":
                return self.config.getboolean(section, key)
            case "str" | "string":
                return self.config.get(section, key)
            case _:
                logger.error("Unknown type {!r}", typ)
                return None

    def set(self, section, key, value):
        """Set config value."""
        self.config.set(section, key, str(value))

    def read(self, paths=None):
        """Read the config from file."""
        if paths is None:
            paths = self.readpaths

        if paths:
            usedfile = self.config.read(paths)
            if not usedfile:
                logger.warning("No valid config file found in {!r}", paths)
        else:
            logger.warning("No file to read from was specified.")

    def write(self, path=None):
        """Write the config to file."""
        if path is None:
            path = self.writepath

        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "w") as cf:
            self.config.write(cf)


Config = ConfigManager(config_paths)



from qtpy import QtGui
from qtpy import QtCore
from qtpy import QtWidgets

import collections
import os
import platform
import subprocess
import json

from typstwriter import logging
from typstwriter import configuration
from typstwriter import globalstate

logger = logging.getLogger(__name__)
config = configuration.Config
state = globalstate.State


class FileIconProvider(QtWidgets.QFileIconProvider):
    """Adds a typst icon to the default QFileIconProvider."""

    def icon(self, info):
        """Return icon associated with info."""
        if isinstance(info, QtCore.QFileInfo):  # noqa SIM102
            if info.suffix() == "typ":
                return QtGui.QIcon(icon_path("typst.png"))
        return super().icon(info)


class RecentFilesModel(QtCore.QAbstractListModel):
    """A data model of the recently used files."""

    def __init__(self, recent_files):
        """Initialize and store data."""
        QtCore.QAbstractListModel.__init__(self)
        self.recent_files = recent_files

    def data(self, index, role):
        """Return the filepath (or its icon) stored under a given index."""
        path = self.recent_files[index.row()]

        match role:
            case QtCore.Qt.DisplayRole:
                return path
            case QtCore.Qt.DecorationRole:
                return FileIconProvider().icon(QtCore.QFileInfo(path))
            case _:
                return None

    def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):  # This is an overriding function # noqa: N802
        """Return the header name."""
        return "Filepath"

    def rowCount(self, index):  # This is an overriding function # noqa: N802
        """Return the number of stored filepaths."""
        return len(self.recent_files)


class RecentFiles(QtCore.QObject):
    """Stores, reads and writes a fixed-size list of recently used files."""

    recent_files_changed = QtCore.Signal(list)

    def __init__(self, path=None):
        """
        Initialize and load recent files list.

        The path of the file storing the list of recent files is set to path argument if given, otherwise use the config value.
        """
        QtCore.QObject.__init__(self)

        maxlen = config.get("Internals", "recent_files_length", "int")
        self.recent_files = collections.deque(maxlen=maxlen)

        if not path:
            path = config.get("Internals", "recent_files_path")

        self.path = os.path.expanduser(path)
        self.read()

    def append(self, file):
        """Append file to the recent files list."""
        if file not in self.recent_files:
            self.recent_files.appendleft(file)
        else:
            self.recent_files.remove(file)
            self.recent_files.appendleft(file)

        self.recent_files_changed.emit(list(self.recent_files))

    def clear(self):
        """Clear the recent files list."""
        self.recent_files.clear()
        self.recent_files_changed.emit(list(self.recent_files))

    def list(self):
        """Return the recent files list."""
        return list(self.recent_files)

    def read(self):
        """Read the list of recently used files from disk."""
        self.clear()
        try:
            with open(self.path, "r") as f:
                for line in f:
                    self.recent_files.append(line.strip())
        except OSError:
            logger.info("Could not read file {!r}.", self.path)

    def write(self):
        """Write the list of recently used files to disk."""
        try:
            os.makedirs(os.path.dirname(self.path), exist_ok=True)
            with open(self.path, "w") as f:
                for line in self.recent_files:
                    f.write(line + "\n")
        except OSError:
            logger.info("Could not write file {!r}.", self.path)


def open_with_external_program(path):
    """Open a file in an external program."""
    if os.path.exists(path):
        match platform.system():
            case "Linux":
                subprocess.call(["xdg-open", path])
            case "Windows":
                os.startfile(path)
            case "Darwin":
                subprocess.call(["open", path])
            case _:
                logger.error("Unsupported system : {}.", platform.system())
    else:
        logger.warning("Attempted to open file with external program but {!r} is not a valid path.", path)


def pdf_path(typst_path):
    """Convert a input path to a pdf output path."""
    (trunk, ext) = os.path.splitext(typst_path)
    pdf_path = trunk + ".pdf"
    return pdf_path


class TogglingAction(QtWidgets.QAction):
    """
    A QAction that can be toggled.

    It automatically changes text and icon when being toggled.
    User triggering causes the activated/deactivated signals being emitted.
    """

    activated = QtCore.Signal()
    deactivated = QtCore.Signal()

    def __init__(self, parent=None):
        """Init."""
        QtWidgets.QAction.__init__(self, parent)
        self.setCheckable(True)
        self.icon_on = QtGui.QIcon()
        self.icon_off = QtGui.QIcon()
        self.text_on = ""
        self.text_off = ""
        self.toggled.connect(self.update_text)  # User or progammatic interaction
        self.toggled.connect(self.update_icon)  # User or progammatic interaction
        self.triggered.connect(self.handle_triggered)  # User interaction

    def setIcon(self, icon, state=None):  # noqa N802
        """Extend parent setIcon with state information."""
        if state == QtGui.QIcon.State.On or state is None:
            self.icon_on = icon
        if state == QtGui.QIcon.State.Off or state is None:
            self.icon_off = icon
        self.update_icon()

    def setText(self, text, state=None):  # noqa N802
        """Extend parent setText with state information."""
        if state == QtGui.QIcon.State.On or state is None:
            self.text_on = text
        if state == QtGui.QIcon.State.Off or state is None:
            self.text_off = text
        self.update_text()

    def update_icon(self):
        """Update icon when toggling."""
        if self.isChecked() is True:
            super().setIcon(self.icon_on)
        else:
            super().setIcon(self.icon_off)

    def update_text(self):
        """Update text when toggling."""
        if self.isChecked() is True:
            super().setText(self.text_on)
        else:
            super().setText(self.text_off)

    def handle_triggered(self, checked):
        """Handle toggling."""
        self.update_text()
        self.update_icon()
        if checked:
            self.activated.emit()
        else:
            self.deactivated.emit()


def icon_path(name):
    """Get an icon by filename."""
    packagedir = os.path.dirname(__file__)
    path = os.path.join(packagedir, "icons", name)
    return path


def selection_start_block(cursor):
    """Return the block number of the selection start."""
    c = QtGui.QTextCursor(cursor)
    c.setPosition(cursor.selectionStart())
    return c.blockNumber()


def selection_end_block(cursor):
    """Return the block number of the selection end."""
    c = QtGui.QTextCursor(cursor)
    c.setPosition(cursor.selectionEnd())
    return c.blockNumber()


def typst_available():
    """Check if typst is available."""
    valid_output = "Typst"
    valid_exit_codes = [0, 2]

    process = QtCore.QProcess()
    process.setProcessChannelMode(QtCore.QProcess.ProcessChannelMode.MergedChannels)
    process.start(config.get("Compiler", "name"))
    process.waitForFinished(msecs=100)
    output = bytes(process.readAll()).decode("utf8")

    return (
        process.exitStatus() == QtCore.QProcess.NormalExit and process.exitCode() in valid_exit_codes and valid_output in output
    )


def qstring_length(text):
    """
    Compute the length of a utf16-encoded QString.

    This function was adopted from spyder: https://github.com/spyder-ide/spyder/blob/f3a0a762f81bf3bf78ed5b18a1f0dc97966249b6/spyder/utils/qstringhelpers.py
    Original Copyright (c) Spyder Project Contributors, originally licensed under the MIT License.
    """
    utf16_text = text.encode("utf16")
    length = len(utf16_text) // 2
    # Remove Byte order mark.
    # TODO: All unicode Non-characters should be removed
    if utf16_text[:2] in [b"\xff\xfe", b"\xff\xff", b"\xfe\xff"]:
        length -= 1
    return length


def read_session_file():
    """Read Session file."""
    path = os.path.expanduser(config.get("Internals", "session_path"))
    try:
        with open(path, "r") as f:
            (working_directory, files) = json.load(f)
        return (working_directory, files)
    except OSError:
        logger.info("Could not read file {!r}.", path)
        return None
    except ValueError:
        logger.warning("Invalid session file {!r}.", path)
        return None


def write_session_file(working_directory, files):
    """Write Session file."""
    path = os.path.expanduser(config.get("Internals", "session_path"))
    try:
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "w") as f:
            json.dump((working_directory, files), f)
    except OSError:
        logger.info("Could not write file {!r}.", path)



# This file is in part derived from the Qt for Python PDF Viewer Example
# (see https://doc.qt.io/qtforpython-6/examples/example_pdfwidgets_pdfviewer.html)
# Original Copyright (C) 2022 The Qt Company Ltd., originally licensed under BSD-3-Clause

from qtpy import QtPdf
from qtpy import QtPdfWidgets
from qtpy import QtWidgets
from qtpy import QtCore
from qtpy import QtGui

import os
from time import time

from typstwriter import util

from typstwriter import logging
from typstwriter import configuration
from typstwriter import globalstate

logger = logging.getLogger(__name__)
config = configuration.Config
state = globalstate.State


class ZoomSelector(QtWidgets.QComboBox):
    """ComboBox for zoom to numerical values or height/width fit."""

    zoom_mode_changed = QtCore.Signal(QtPdfWidgets.QPdfView.ZoomMode)
    zoom_factor_changed = QtCore.Signal(float)

    def __init__(self, parent):
        """Init and set zoom levels."""
        super().__init__(parent)
        self.setEditable(True)

        self.addItem("Fit Page Width")
        self.addItem("Fit Page Height")
        self.insertSeparator(2)
        self.addItem("12%")
        self.addItem("25%")
        self.addItem("33%")
        self.addItem("50%")
        self.addItem("66%")
        self.addItem("75%")
        self.addItem("100%")
        self.addItem("125%")
        self.addItem("150%")
        self.addItem("200%")
        self.addItem("400%")

        self.currentTextChanged.connect(self.on_current_text_changed)
        self.lineEdit().editingFinished.connect(self._editing_finished)

    @QtCore.Slot(float)
    def set_zoom_factor(self, zoom_factor):
        """Set Zoom."""
        percent = int(zoom_factor * 100)
        self.setCurrentText(f"{percent}%")

    @QtCore.Slot()
    def reset(self):
        """Reset Zoom."""
        self.setCurrentText("100%")

    @QtCore.Slot(str)
    def on_current_text_changed(self, text):
        """Handle change in ComboBox and set Zoom."""
        if text == "Fit Page Width":
            self.zoom_mode_changed.emit(QtPdfWidgets.QPdfView.ZoomMode.FitToWidth)
        elif text == "Fit Page Height":
            self.zoom_mode_changed.emit(QtPdfWidgets.QPdfView.ZoomMode.FitInView)
        elif text.endswith("%"):
            factor = 1.0
            zoom_level = int(text[:-1])
            factor = zoom_level / 100.0
            self.zoom_mode_changed.emit(QtPdfWidgets.QPdfView.ZoomMode.Custom)
            self.zoom_factor_changed.emit(factor)

    @QtCore.Slot()
    def _editing_finished(self):
        """Relay zoom level to on_current_text_changed."""
        self.on_current_text_changed(self.lineEdit().text())


class PDFViewer(QtWidgets.QFrame):
    """A PDF Viewer."""

    def __init__(self):
        """Populate the PDF Viewer and create document store."""
        QtWidgets.QFrame.__init__(self)
        self.setFrameStyle(QtWidgets.QFrame.Shape.StyledPanel)

        # Action Zoom In
        self.actionZoom_In = QtGui.QAction(self)
        self.actionZoom_In.setIcon(QtGui.QIcon.fromTheme(QtGui.QIcon.ZoomIn, QtGui.QIcon(util.icon_path("plus.svg"))))
        self.actionZoom_In.setText("Zoom In")
        self.actionZoom_In.setShortcut(QtGui.QKeySequence.ZoomIn)

        # Action Zoom Out
        self.actionZoom_Out = QtGui.QAction(self)
        self.actionZoom_Out.setIcon(QtGui.QIcon.fromTheme(QtGui.QIcon.ZoomOut, QtGui.QIcon(util.icon_path("minus.svg"))))
        self.actionZoom_Out.setText("Zoom Out")
        self.actionZoom_Out.setShortcut(QtGui.QKeySequence.ZoomOut)

        # Action Previous Page
        self.actionPrevious_Page = QtGui.QAction(self)
        self.actionPrevious_Page.setIcon(
            QtGui.QIcon.fromTheme(QtGui.QIcon.GoPrevious, QtGui.QIcon(util.icon_path("larrow.svg")))
        )
        self.actionPrevious_Page.setText("Previous Page")

        # Action Next Page
        self.actionNext_Page = QtGui.QAction(self)
        self.actionNext_Page.setIcon(QtGui.QIcon.fromTheme(QtGui.QIcon.GoNext, QtGui.QIcon(util.icon_path("rarrow.svg"))))
        self.actionNext_Page.setText("Next Page")
        self.actionPrevious_Page.setShortcut("PgUp")

        # Action External Viewer
        self.actionOpen_External = QtGui.QAction(self)
        self.actionOpen_External.setIcon(QtGui.QIcon.fromTheme("viewpdf", QtGui.QIcon(util.icon_path("pdf.svg"))))
        self.actionOpen_External.setText("External Viewer")
        self.actionNext_Page.setShortcut("PgDown")

        # Layout
        self.verticalLayout = QtWidgets.QVBoxLayout(self)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)

        # Toolbar
        self.toolbar = QtWidgets.QToolBar(self)
        self.toolbar.setMovable(False)
        self.toolbar.setFloatable(False)
        self.verticalLayout.addWidget(self.toolbar)

        # PDF view
        self.pdfView = QtPdfWidgets.QPdfView(self)
        size_policy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        size_policy.setHorizontalStretch(10)
        size_policy.setVerticalStretch(0)
        size_policy.setHeightForWidth(self.pdfView.sizePolicy().hasHeightForWidth())
        self.pdfView.setSizePolicy(size_policy)
        self.verticalLayout.addWidget(self.pdfView)

        # Connect Signals and Slots
        self.actionZoom_In.triggered.connect(self.zoom_in_triggered)
        self.actionZoom_Out.triggered.connect(self.zoom_out_triggered)
        self.actionPrevious_Page.triggered.connect(self.previous_page_triggered)
        self.actionNext_Page.triggered.connect(self.next_page_triggered)
        self.actionOpen_External.triggered.connect(self.open_in_external_viewer)

        self.m_zoomSelector = ZoomSelector(self)
        self.m_pageSelector = QtWidgets.QSpinBox(self)
        self.m_maxPage = QtWidgets.QLabel(self)
        self.m_document = QtPdf.QPdfDocument(self)

        self.m_zoomSelector.setMaximumWidth(150)

        self.toolbarspacer1 = QtWidgets.QWidget()
        self.toolbarspacer1.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        self.toolbarspacer2 = QtWidgets.QWidget()
        self.toolbarspacer2.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)

        # Populate toolbar
        self.toolbar.addAction(self.actionZoom_Out)
        self.toolbar.addWidget(self.m_zoomSelector)
        self.toolbar.addAction(self.actionZoom_In)
        self.toolbar.addWidget(self.toolbarspacer1)
        self.toolbar.addAction(self.actionPrevious_Page)
        self.toolbar.addWidget(self.m_pageSelector)
        self.toolbar.addWidget(self.m_maxPage)
        self.toolbar.addAction(self.actionNext_Page)
        self.toolbar.addWidget(self.toolbarspacer2)
        self.toolbar.addAction(self.actionOpen_External)

        self.m_pageSelector.valueChanged.connect(lambda n: self.page_selected(n - 1))
        nav = self.pdfView.pageNavigator()
        nav.currentPageChanged.connect(lambda n: self.m_pageSelector.setValue(n + 1))
        self.m_pageSelector.setMinimum(1)
        # nav.backAvailableChanged.connect(self.actionBack.setEnabled)
        # nav.forwardAvailableChanged.connect(self.actionForward.setEnabled)

        self.m_maxPage.setText(f" of {self.m_document.pageCount()}")

        self.m_zoomSelector.zoom_mode_changed.connect(self.pdfView.setZoomMode)
        self.m_zoomSelector.zoom_factor_changed.connect(self.pdfView.setZoomFactor)
        self.m_zoomSelector.reset()

        self.pdfView.setDocument(self.m_document)
        self.pdfView.zoomFactorChanged.connect(self.m_zoomSelector.set_zoom_factor)
        self.pdfView.setPageMode(QtPdfWidgets.QPdfView.PageMode.MultiPage)

        self.docpath = None

    @QtCore.Slot(QtCore.QUrl)
    def open(self, path):
        """Open and display a file."""
        self.docpath = path
        if os.path.exists(path):
            self.m_document.load(path)
            self.page_selected(0)
            self.m_pageSelector.setMaximum(self.m_document.pageCount())
            self.m_maxPage.setText(f" of {self.m_document.pageCount()}")
        else:
            self.m_document.close()
            logger.warning("{!r} is not a valid file", path)

    @QtCore.Slot()
    def reload(self):
        """Reload file."""
        if self.docpath and os.path.exists(self.docpath):
            # get current scroll position
            pos = self.pdfView.verticalScrollBar().value()

            # reload
            t1 = time()
            self.m_document.load(self.docpath)
            t2 = time()

            # update pages
            self.m_pageSelector.setMaximum(self.m_document.pageCount())
            self.m_maxPage.setText(f" of {self.m_document.pageCount()}")

            # scroll back to same point
            self.pdfView.verticalScrollBar().setValue(pos)

            logger.debug("Reloaded from disk in {:.2f}ms.", (t2 - t1) * 1000)
        else:
            logger.debug("Attempted to reload PDF but no valid file found at {!r}.", self.docpath)

    @QtCore.Slot(int)
    def page_selected(self, page):
        """Jump to page."""
        nav = self.pdfView.pageNavigator()
        nav.jump(page, QtCore.QPointF(), nav.currentZoom())

    @QtCore.Slot()
    def zoom_in_triggered(self):
        """Zoom in."""
        factor = self.pdfView.zoomFactor() * 1.2
        self.pdfView.setZoomFactor(factor)

    @QtCore.Slot()
    def zoom_out_triggered(self):
        """Zoom out."""
        factor = self.pdfView.zoomFactor() / 1.2
        self.pdfView.setZoomFactor(factor)

    @QtCore.Slot()
    def previous_page_triggered(self):
        """Switch to previous page."""
        nav = self.pdfView.pageNavigator()
        nav.jump(max(nav.currentPage() - 1, 0), QtCore.QPointF(), nav.currentZoom())

    @QtCore.Slot()
    def next_page_triggered(self):
        """Switch to next page."""
        nav = self.pdfView.pageNavigator()
        nav.jump(min(nav.currentPage() + 1, self.pdfView.document().pageCount() - 1), QtCore.QPointF(), nav.currentZoom())

    @QtCore.Slot()
    def open_in_external_viewer(self):
        """Open the pdf file in an external viewer."""
        if self.docpath:
            util.open_with_external_program(self.docpath)
        else:
            logger.warning("Attempted to open external PDF viewer but no file is loaded.")



import logging
# See https://python.readthedocs.io/en/stable/howto/logging-cookbook.html#use-of-alternative-formatting-styles


class Message:
    """A logger message."""

    def __init__(self, fmt, args):
        """Init."""
        self.fmt = fmt
        self.args = args

    def __str__(self):
        """Format string."""
        return self.fmt.format(*self.args)


class StyleAdapter(logging.LoggerAdapter):
    """Curly brace style adapter."""

    def __init__(self, logger, extra=None):
        """Init."""
        super().__init__(logger, extra or {})

    def log(self, level, msg, *args, **kwargs):
        """Log if enabled."""
        if self.isEnabledFor(level):
            msg, kwargs = self.process(msg, kwargs)
            self.logger._log(level, Message(msg, args), (), stacklevel=2, **kwargs)


def setup_logger(level):
    """Set the main logger up."""
    fmt = "{levelname:8} {asctime:26} {filename:20} {funcName:26} line {lineno:<4d}: {message}"
    formatter = logging.Formatter(fmt=fmt, style="{")

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    log = logging.root
    log.addHandler(handler)
    if level in ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]:
        log.setLevel(level)
    else:
        log.setLevel("INFO")


def getLogger(name):  # noqa N802
    """Get brace style logger."""
    return StyleAdapter(logging.getLogger(name))



from qtpy import QtGui
from qtpy import QtCore
from qtpy import QtWidgets

from typstwriter import util

from typstwriter import logging
from typstwriter import configuration
from typstwriter import globalstate

logger = logging.getLogger(__name__)
config = configuration.Config
state = globalstate.State


class MenuBar(QtWidgets.QMenuBar):
    """Main Window Menu Bar."""

    def __init__(self, actions):
        """Populate menu bar."""
        QtWidgets.QMenuBar.__init__(self)

        self.menuFile = QtWidgets.QMenu(self)
        self.menuEdit = QtWidgets.QMenu(self)
        self.menuView = QtWidgets.QMenu(self)
        self.menuSettings = QtWidgets.QMenu(self)

        self.menuFile.setTitle("File")
        self.menuEdit.setTitle("Edit")
        self.menuView.setTitle("View")
        self.menuSettings.setTitle("Settings")

        self.addAction(self.menuFile.menuAction())
        self.addAction(self.menuEdit.menuAction())
        self.addAction(self.menuView.menuAction())
        self.addAction(self.menuSettings.menuAction())

        self.menuFile.addAction(actions.new_File)
        self.menuFile.addSeparator()
        self.menuFile.addAction(actions.open_File)
        self.recent_files_menu = RecentFilesMenu()
        self.menuFile.addMenu(self.recent_files_menu)
        self.menuFile.addAction(actions.load_last_Session)
        self.menuFile.addSeparator()
        self.menuFile.addAction(actions.save)
        self.menuFile.addAction(actions.save_as)
        self.menuFile.addSeparator()
        self.menuFile.addAction(actions.close)
        self.menuFile.addSeparator()
        self.menuFile.addAction(actions.quit)
        self.menuFile.addSeparator()

        self.menuEdit.addAction(actions.copy)
        self.menuEdit.addAction(actions.cut)
        self.menuEdit.addAction(actions.paste)
        self.menuEdit.addSeparator()
        self.menuEdit.addAction(actions.search)

        self.layout_menu = QtWidgets.QMenu(self)
        self.layout_menu.setTitle("Layout")
        self.layout_menu.addActions(actions.layout.actions())

        self.editor_zoom_menu = QtWidgets.QMenu(self)
        self.editor_zoom_menu.setTitle("Editor Zoom")
        self.editor_zoom_menu.addAction(actions.font_size_up)
        self.editor_zoom_menu.addAction(actions.font_size_dn)
        self.editor_zoom_menu.addAction(actions.font_size_reset)

        self.menuView.addMenu(self.layout_menu)
        self.menuView.addSeparator()
        self.menuView.addAction(actions.show_fs_explorer)
        self.menuView.addAction(actions.show_compiler_options)
        self.menuView.addAction(actions.show_compiler_output)
        self.menuView.addSeparator()
        self.menuView.addMenu(self.editor_zoom_menu)

        self.menuSettings.addAction(actions.open_config)


class RecentFilesMenu(QtWidgets.QMenu):
    """Menu showing the recently used files."""

    open_file = QtCore.Signal(str)

    def __init__(self):
        """Init."""
        QtWidgets.QMenu.__init__(self)

        self.setIcon(QtGui.QIcon.fromTheme(QtGui.QIcon.DocumentOpenRecent, QtGui.QIcon(util.icon_path("recentFile.svg"))))
        self.setTitle("Open recent file")

        self.addAction("")

        self.triggered.connect(self.open_from_action)

    def display_recent_files(self, recent_files):
        """Diplay a list of recent files."""
        self.clear()
        for f in recent_files:
            self.addAction(f)

    def open_from_action(self, action):
        """Emit open_file signal with text of given action."""
        self.open_file.emit(action.text())



# This file is in part derived from superqt CodeSyntaxHighlight
# (see https://github.com/pyapp-kit/superqt/blob/ac4adf523442e14049a56c012eeb23d0c2c3d314/src/superqt/utils/_code_syntax_highlight.py)
# Original Copyright (c) 2021, Talley Lambert, originally licensed under BSD-3-Clause

from qtpy import QtGui

import os
import collections

import pygments
from pygments import formatter
from pygments import lexers

from typstwriter import util

from typstwriter import logging
from typstwriter import configuration
from typstwriter import globalstate

logger = logging.getLogger(__name__)
config = configuration.Config
state = globalstate.State


def get_lexer_by_name(name):
    """Get a lexer by name, if not found return “Null” lexer."""
    for lexer_name, _, _, _ in lexers.get_all_lexers(plugins=False):
        if lexer_name == name:
            return lexers.find_lexer_class(name)()
    return lexers.TextLexer()


def get_lexer_name_by_filename(path):
    """
    Get the name of a lexer for a filename.

    This is a simplified(but faster) alternative to pygments get_lexer_for_filename.
    """
    if not path:
        return None

    filename = os.path.basename(path)
    for name, _, filenames, _ in lexers.get_all_lexers(plugins=False):
        for f in filenames:
            if lexers._fn_matches(filename, f):
                return name
    return None


def available_lexers():
    """Return a list with the names of all available lexers."""
    return [name for name, _, _, _ in lexers.get_all_lexers(plugins=False)]


class QTextCharFormatter(formatter.Formatter):
    """pygments formatter using QTextCharFormat."""

    def __init__(self, **kwargs):
        """Init."""
        super().__init__(**kwargs)
        self.token_map = collections.defaultdict(QtGui.QTextCharFormat)
        for token, _ in self.style:
            self.token_map[token] = self.get_format(token)

    def get_format(self, token):
        """Get QTextCharFormat for a token."""
        style = self.style.style_for_token(token)

        text_char_format = QtGui.QTextCharFormat()
        text_char_format.setFontFamilies(["monospace"])
        if style.get("color"):
            text_char_format.setForeground(QtGui.QColor(f"#{style['color']}"))
        if style.get("bgcolor"):
            text_char_format.setBackground(QtGui.QColor(f"#{style['bgcolor']}"))
        if style.get("bold"):
            text_char_format.setFontWeight(QtGui.QFont.Bold)
        if style.get("italic"):
            text_char_format.setFontItalic(True)
        if style.get("underline"):
            text_char_format.setFontUnderline(True)
        if style.get("border"):
            # TODO: implement border highlighting
            pass

        return text_char_format

    def format(self, tokensource, outfile):
        """
        Format the given token stream.

        `outfile` is needed from parent class, but is unused.
        """
        for token, value in tokensource:
            yield (self.token_map[token], util.qstring_length(value))


class CodeSyntaxHighlight(QtGui.QSyntaxHighlighter):
    """Generic syntax highlighter making use of pygments."""

    def __init__(self, parent, lexer, theme):
        """Init."""
        super().__init__(parent)
        self.formatter = QTextCharFormatter(style=theme)
        self.lexer = lexer

    @property
    def font_color(self):
        """Font color."""
        return self.formatter.style.styles[pygments.token.Token]

    @property
    def background_color(self):
        """Background color."""
        return self.formatter.style.background_color

    @property
    def highlight_color(self):
        """Highlight color."""
        return self.formatter.style.highlight_color

    @property
    def line_number_color(self):
        """Line number color."""
        return self.formatter.style.line_number_color

    @property
    def line_number_background_color(self):
        """Line number background color."""
        return self.formatter.style.line_number_background_color

    @property
    def line_number_special_color(self):
        """Line number special color."""
        return self.formatter.style.line_number_special_color

    @property
    def line_number_special_background_color(self):
        """Line number special background color."""
        return self.formatter.style.line_number_special_background_color

    def highlightBlock(self, text):  # This is an overriding function # noqa: N802
        """Highlight the given text block."""
        format_list = self.formatter.format(pygments.lex(text, self.lexer), None)
        start = 0
        for format, length in format_list:
            self.setFormat(start, length, format)
            start += length



from qtpy import QtWidgets

from typstwriter import fs_explorer


class TestFSExplorer:
    """Test fs_explorer.FSExplorer."""

    def test_open_directory_dialog(self, tmp_path, qtbot, monkeypatch):
        """Make sure the root can be set."""
        fse = fs_explorer.FSExplorer()
        qtbot.addWidget(fse)

        existant_path = tmp_path / "test_folder_existant"
        existant_path.mkdir()
        monkeypatch.setattr(QtWidgets.QFileDialog, "getExistingDirectory", lambda *args: str(existant_path))
        fse.open_directory_dialog()
        assert fse.root == str(existant_path)

        non_existant_path = tmp_path / "test_folder_non_existant"
        monkeypatch.setattr(QtWidgets.QFileDialog, "getExistingDirectory", lambda *args: str(non_existant_path))
        fse.open_directory_dialog()
        assert fse.root != str(non_existant_path)

    def test_root_setting(self, tmp_path, qtbot):
        """Make sure the root can be set."""
        fse = fs_explorer.FSExplorer()
        qtbot.addWidget(fse)
        fse.set_root(str(tmp_path))

        assert fse.root == str(tmp_path)
        assert fse.filesystem_model.rootPath() == str(tmp_path)
        assert fse.completer_filesystem_model.rootPath() == str(tmp_path)
        assert fse.pathBar.text() == str(tmp_path)
        assert fs_explorer.state.working_directory.Value == str(tmp_path)

    def test_line_edited(self, tmp_path, qtbot):
        """Make sure line_edited changes root if valid."""
        fse = fs_explorer.FSExplorer()
        qtbot.addWidget(fse)

        # Valid path
        fse.pathBar.setText(str(tmp_path))
        fse.pathBar.editingFinished.emit()
        assert fse.root == str(tmp_path)

        # Invalid path
        text = fse.pathBar.text()
        root = fse.root
        fse.pathBar.setText(str(tmp_path / "nonexistent_dir"))
        fse.pathBar.editingFinished.emit()
        assert fse.root == root
        assert fse.pathBar.text() == text

    def test_goto_parent_directory(self, tmp_path, qtbot):
        """Make sure the root can be set."""
        fse = fs_explorer.FSExplorer()
        qtbot.addWidget(fse)

        path = tmp_path / "test_path"
        path.mkdir()
        fse.set_root(str(path))
        fse.goto_parent_directory()
        assert fse.root == str(tmp_path)

    def test_new_file(self, tmp_path, qtbot, caplog, monkeypatch):
        """Make sure a new file behaves correctly."""
        fse = fs_explorer.FSExplorer()
        qtbot.addWidget(fse)
        monkeypatch.setattr(QtWidgets.QMessageBox, "warning", lambda *args: QtWidgets.QMessageBox.Ok)

        file_name = "test_file.dat"
        file_path = tmp_path / file_name

        # Automatically answer dialog
        monkeypatch.setattr(QtWidgets.QInputDialog, "getText", lambda *args: (file_name, True))

        # Make sure the file can be created
        assert not file_path.exists()
        fse.new_file_in_dir(str(tmp_path))
        assert file_path.exists()

        # Make sure the file is not overwritten
        file_path.write_text("Test")
        assert file_path.read_text() == "Test"
        fse.new_file_in_dir(str(tmp_path))
        assert "already exists" in caplog.text
        assert file_path.read_text() == "Test"

        # Make sure the file is overwritten if flag is set
        file_path.write_text("Test")
        assert file_path.read_text() == "Test"
        fse.new_file(str(file_path), overwrite=True)
        assert file_path.read_text() != "Test"

    def test_new_folder(self, tmp_path, qtbot, caplog, monkeypatch):
        """Make sure a new folder behaves correctly."""
        fse = fs_explorer.FSExplorer()
        qtbot.addWidget(fse)
        monkeypatch.setattr(QtWidgets.QMessageBox, "warning", lambda *args: QtWidgets.QMessageBox.Ok)

        folder_name = "test_folder"
        folder_path = tmp_path / folder_name

        # Automatically answer dialog
        monkeypatch.setattr(QtWidgets.QInputDialog, "getText", lambda *args: (folder_name, True))

        # Make sure the folder can be created
        assert not folder_path.exists()
        fse.new_folder_in_dir(str(tmp_path))
        assert folder_path.exists()

        # Make sure the folder is not overwritten
        assert folder_path.exists()
        fse.new_folder_in_dir(str(tmp_path))
        assert "already exists" in caplog.text

    def test_rename(self, tmp_path, qtbot, caplog, monkeypatch):
        """Make sure rename works."""
        fse = fs_explorer.FSExplorer()
        qtbot.addWidget(fse)
        monkeypatch.setattr(QtWidgets.QMessageBox, "warning", lambda *args: QtWidgets.QMessageBox.Ok)

        from_name = "test_file_from"
        to_name = "test_file_to"
        from_path = tmp_path / from_name
        to_path = tmp_path / to_name
        from_path.touch()

        # Automatically answer dialog
        monkeypatch.setattr(QtWidgets.QInputDialog, "getText", lambda *args, **kwargs: (to_name, True))

        # Make sure the file can be renamed
        assert from_path.exists()
        assert not to_path.exists()
        fse.rename_from(str(from_path))
        assert not from_path.exists()
        assert to_path.exists()

        # Make sure the file is not overwritten
        from_path.touch()
        assert from_path.exists()
        assert to_path.exists()
        fse.rename_from(str(from_path))
        assert "already exists" in caplog.text
        assert from_path.exists()
        assert to_path.exists()

    def test_copy_from_to(self, tmp_path, qtbot, caplog, monkeypatch):
        """Make sure copying works."""
        fse = fs_explorer.FSExplorer()
        qtbot.addWidget(fse)
        monkeypatch.setattr(QtWidgets.QMessageBox, "warning", lambda *args: QtWidgets.QMessageBox.Ok)

        name = "test_file_name"
        from_dir = tmp_path / "from"
        to_dir = tmp_path / "to"
        from_dir.mkdir()
        to_dir.mkdir()
        from_path = from_dir / name
        to_path = to_dir / name

        from_path.touch()

        # Make sure the file can be copied
        assert from_path.exists()
        assert not to_path.exists()
        fse.copy_from_to(str(from_path), str(to_path))
        assert from_path.exists()
        assert to_path.exists()

        # Make sure the file is not overwritten
        from_path.touch()
        to_path.touch()
        assert from_path.exists()
        assert to_path.exists()
        fse.copy_from_to(str(from_path), str(to_path))
        assert "already exists" in caplog.text
        assert from_path.exists()
        assert to_path.exists()

    def test_delete(self, tmp_path, qtbot, monkeypatch):
        """Make sure a folder or file can be deleted."""
        fse = fs_explorer.FSExplorer()
        qtbot.addWidget(fse)

        folder_path = tmp_path / "test_folder"
        file_path = tmp_path / "test_file.dat"

        folder_path.mkdir()
        file_path.touch()
        assert file_path.exists()
        assert folder_path.exists()

        # Decline deletion
        monkeypatch.setattr(QtWidgets.QMessageBox, "question", lambda *args: QtWidgets.QMessageBox.StandardButton.No)
        fse.delete(str(folder_path))
        fse.delete(str(file_path))
        assert file_path.exists()
        assert folder_path.exists()

        # Accept deletion
        monkeypatch.setattr(QtWidgets.QMessageBox, "question", lambda *args: QtWidgets.QMessageBox.StandardButton.Yes)
        fse.delete(str(folder_path))
        fse.delete(str(file_path))
        assert not file_path.exists()
        assert not folder_path.exists()



import pytest

import configparser

from typstwriter import configuration

example_config = """
[Section_1]
key_1 = /home/user/Documents/

[Section_2]
key_1 = 200
key_2 = Some Text
key_3 = True
key_4 = 3.5

[Compiler]
name = custom_typst_cli
mode = on_demand
"""


@pytest.fixture()
def custom_config(tmp_path):
    """Write a custom config file and return the corresponding path."""
    config_file_path = tmp_path / "config_file.ini"
    config_file_path.write_text(example_config)
    return config_file_path


class TestConfigManager:
    """Test configuration.ConfigManager."""

    def test_constructor_default_config(self, tmp_path):
        """Make sure the constructor correctly loads the default config."""
        cm = configuration.ConfigManager()

        assert cm.config.get("Compiler", "name") == "typst"
        assert cm.writepath == str(configuration.default_config_path)

    def test_constructor_custom_config(self, custom_config):
        """Make sure the constructor correctly loads a config file."""
        cm = configuration.ConfigManager(custom_config)

        assert cm.config.get("Compiler", "name") == "custom_typst_cli"
        assert cm.writepath == str(custom_config)

    def test_get(self, custom_config, caplog):
        """Test the get function."""
        cm = configuration.ConfigManager(custom_config)

        assert cm.get("Section_2", "key_1") == "200"
        assert cm.get("Section_2", "key_2", typ="string") == "Some Text"
        assert cm.get("Section_2", "key_1", typ="int") == 200  # noqa: PLR2004
        assert cm.get("Section_2", "key_4", typ="float") == 3.5  # noqa: PLR2004
        assert cm.get("Section_2", "key_3", typ="bool") is True
        assert cm.get("Section_2", "key_3", typ="no_type") is None
        assert "Unknown type" in caplog.text

        with pytest.raises(configparser.NoOptionError):
            cm.get("Section_2", "no_such_key")

    def test_set(self, custom_config):
        """Test the set function."""
        cm = configuration.ConfigManager(custom_config)

        cm.set("Section_2", "key_1", 44)
        assert cm.config.get("Section_2", "key_1") == "44"

        cm.set("Section_2", "key_1", "44")
        assert cm.config.get("Section_2", "key_1") == "44"

        cm.set("Section_2", "key_1", False)
        assert cm.config.get("Section_2", "key_1") == "False"

    def test_read_default(self, custom_config, caplog):
        """Test the read function with no path provided."""
        # Initially no file provided
        cm = configuration.ConfigManager()
        cm.read()
        assert "No file to read from was specified." in caplog.text

        # Initial file provided
        cm = configuration.ConfigManager(custom_config)
        cm.set("Compiler", "name", "typst")
        assert cm.config.get("Compiler", "name") == "typst"
        cm.read()
        assert cm.config.get("Compiler", "name") == "custom_typst_cli"

    def test_read_path(self, custom_config):
        """Test the read function with path provided."""
        cm = configuration.ConfigManager()
        assert cm.config.get("Compiler", "name") == "typst"
        cm.read(paths=custom_config)
        assert cm.config.get("Compiler", "name") == "custom_typst_cli"

    def test_write_default(self, custom_config):
        """Test the write function with no path provided."""
        cm = configuration.ConfigManager(custom_config)
        custom_config.unlink()
        assert not custom_config.exists()
        cm.write()
        assert custom_config.exists()

    def test_write_path(self, custom_config, tmp_path):
        """Test the write function with path provided."""
        cm = configuration.ConfigManager(custom_config)
        file_out = tmp_path / "folder" / "configfile.ini"
        assert not file_out.exists()
        cm.write(path=file_out)
        assert file_out.exists()



from qtpy import QtGui
from qtpy import QtCore
from qtpy import QtWidgets

from typstwriter import menubar
from typstwriter import toolbar
from typstwriter import actions
from typstwriter import editor
from typstwriter import pdf_viewer
from typstwriter import fs_explorer
from typstwriter import compiler_tools
from typstwriter import compiler
from typstwriter import util

from typstwriter import logging
from typstwriter import configuration
from typstwriter import globalstate
from typstwriter import arguments

logger = logging.getLogger(__name__)
config = configuration.Config
state = globalstate.State
args = arguments.Args


class MainWindow(QtWidgets.QMainWindow):
    """The main window."""

    def __init__(self):
        """Init."""
        super(self.__class__, self).__init__()

        self.setWindowTitle("Typstwriter")
        self.setObjectName("MainWindow")
        self.setWindowIcon(QtGui.QIcon(util.icon_path("typstwriter.svg")))

        # Actions
        self.actions = actions.Actions(self)

        # Menubar
        self.menubar = menubar.MenuBar(self.actions)
        self.setMenuBar(self.menubar)

        # Toolbar
        self.toolbar = toolbar.ToolBar(self.actions)
        self.addToolBar(self.toolbar)

        # Host Widget for Editor and PDF Viewer
        self.WidtetCore = QtWidgets.QWidget(self)
        self.CoreLayout_v = QtWidgets.QVBoxLayout(self.WidtetCore)
        self.CoreLayout_v.setContentsMargins(4, 4, 4, 4)
        self.CoreLayout_v.setSpacing(0)
        self.setCentralWidget(self.WidtetCore)

        self.splitter = QtWidgets.QSplitter(self.WidtetCore)
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        self.CoreLayout_v.addWidget(self.splitter)

        # Host widget for PDF Viewer
        self.Widgetidget_31 = QtWidgets.QWidget(self.splitter)
        self.verticalLayout_1 = QtWidgets.QVBoxLayout(self.Widgetidget_31)
        self.verticalLayout_1.setContentsMargins(0, 0, 0, 0)

        # Host Widget for Editor
        self.Widgetidget = QtWidgets.QWidget(self.splitter)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.Widgetidget)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)

        # PDF Viewer
        self.PDFWidget = pdf_viewer.PDFViewer()
        self.verticalLayout_1.addWidget(self.PDFWidget)

        # Editor
        self.editor = editor.Editor()
        self.verticalLayout_3.addWidget(self.editor)

        # FSExplorer
        self.FSExplorer = fs_explorer.FSExplorer()
        self.FSdock = QtWidgets.QDockWidget("File System Explorer", self)
        self.FSdock.setWidget(self.FSExplorer)
        self.FSdock.setFeatures(QtWidgets.QDockWidget.DockWidgetMovable)
        self.FSdock.setAllowedAreas(QtCore.Qt.DockWidgetArea.LeftDockWidgetArea | QtCore.Qt.DockWidgetArea.RightDockWidgetArea)
        self.addDockWidget(QtCore.Qt.LeftDockWidgetArea, self.FSdock)

        # CompilerOptions
        self.CompilerOptions = compiler_tools.CompilerOptions()
        self.CompilerOptionsdock = QtWidgets.QDockWidget("Compiler Options", self)
        self.CompilerOptionsdock.setWidget(self.CompilerOptions)
        self.CompilerOptionsdock.setFeatures(QtWidgets.QDockWidget.DockWidgetMovable)
        self.CompilerOptionsdock.setAllowedAreas(
            QtCore.Qt.DockWidgetArea.LeftDockWidgetArea | QtCore.Qt.DockWidgetArea.RightDockWidgetArea
        )
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.CompilerOptionsdock)

        # Compiler Output
        self.CompilerOutput = compiler_tools.CompilerOutput()
        self.CompilerOutputdock = QtWidgets.QDockWidget("Compiler Output", self)
        self.CompilerOutputdock.setWidget(self.CompilerOutput)
        self.CompilerOutputdock.setFeatures(QtWidgets.QDockWidget.DockWidgetMovable)
        self.CompilerOutputdock.setAllowedAreas(
            QtCore.Qt.DockWidgetArea.LeftDockWidgetArea | QtCore.Qt.DockWidgetArea.RightDockWidgetArea
        )
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.CompilerOutputdock)

        # CompilerConnector
        self.CompilerConnector = compiler.WrappedCompilerConnector(state.compiler_mode.Value)

        # Connect signals and slots
        self.actions.new_File.triggered.connect(self.editor.new_file)
        self.actions.open_File.triggered.connect(self.editor.open_file_dialog)
        self.editor.recent_files_changed.connect(self.menubar.recent_files_menu.display_recent_files)
        self.menubar.recent_files_menu.open_file.connect(self.editor.open_file)
        self.menubar.recent_files_menu.display_recent_files(self.editor.recentFiles.list())
        self.actions.load_last_Session.triggered.connect(self.load_session)
        self.actions.save.triggered.connect(self.editor.saveactive_tab)
        self.actions.save_as.triggered.connect(self.editor.saveactive_tab_as)
        self.actions.close.triggered.connect(self.editor.closeactive_tab)
        self.actions.quit.triggered.connect(self.close)
        self.actions.copy.triggered.connect(self.editor.copy)
        self.actions.cut.triggered.connect(self.editor.cut)
        self.actions.paste.triggered.connect(self.editor.paste)
        self.actions.search.triggered.connect(self.editor.search)
        self.actions.layout_typewriter.triggered.connect(self.set_layout_typewriter)
        self.actions.layout_editorL.triggered.connect(self.set_layout_editorL)
        self.actions.layout_editorR.triggered.connect(self.set_layout_editorR)
        self.actions.font_size_up.triggered.connect(self.editor.increase_font_size)
        self.actions.font_size_dn.triggered.connect(self.editor.decrease_font_size)
        self.actions.font_size_reset.triggered.connect(self.editor.reset_font_size)
        self.actions.show_fs_explorer.toggled.connect(self.set_fs_explorer_visibility)
        self.actions.show_compiler_options.toggled.connect(self.set_compiler_options_visibility)
        self.actions.show_compiler_output.toggled.connect(self.set_compiler_output_visibility)
        self.actions.show_fs_explorer.setChecked(True)
        self.actions.show_compiler_options.setChecked(True)
        self.actions.show_compiler_output.setChecked(True)
        if config.get("Editor", "save_at_run", "bool"):
            self.actions.run.activated.connect(self.editor.save_all)
        self.actions.run.activated.connect(self.prepare_compilation)
        self.actions.run.activated.connect(self.CompilerConnector.start)
        self.actions.run.deactivated.connect(self.CompilerConnector.stop)
        self.CompilerConnector.started.connect(lambda: self.actions.run.setChecked(True))
        self.CompilerConnector.stopped.connect(lambda: self.actions.run.setChecked(False))
        self.actions.open_config.triggered.connect(self.open_config)

        self.FSExplorer.open_file.connect(self.editor.open_file)
        state.working_directory.Signal.connect(self.FSExplorer.root_changed)
        self.editor.text_changed.connect(self.CompilerConnector.source_changed)
        self.CompilerConnector.document_changed.connect(self.PDFWidget.reload)
        self.CompilerConnector.compilation_finished.connect(self.editor.clear_errors)
        self.CompilerConnector.error_report.connect(self.editor.apply_errors)
        state.main_file.Signal.connect(lambda s: self.CompilerConnector.stop())
        state.main_file.Signal.connect(lambda s: self.CompilerOptions.main_changed(s))
        state.main_file.Signal.connect(lambda s: self.PDFWidget.open(util.pdf_path(s)))

        # For now only display errors
        self.CompilerConnector.compilation_started.connect(self.CompilerOutput.insert_block)
        self.CompilerConnector.new_stderr.connect(self.CompilerOutput.append_to_block)
        state.compiler_mode.Signal.connect(self.CompilerConnector.switch_compiler)

        # Display
        self.showMaximized()

        # Use default layout
        self.use_default_layout()

        # Open files if given as arguments
        for file in args.files:
            self.editor.open_file(file)

        # Check if typst is available
        QtCore.QTimer().singleShot(0, self.check_typst_availability)

        if config.get("General", "resume_last_session", "bool"):
            QtCore.QTimer().singleShot(0, self.load_session)

        logger.info("Gui ready")

    def set_layout_typewriter(self):
        """Set Typstwriter Layout: PDF on top, Editor on bottom."""
        # Vertical orientation
        self.splitter.setOrientation(QtCore.Qt.Vertical)

        # Reorder widgets
        self.splitter.addWidget(self.Widgetidget_31)
        self.splitter.addWidget(self.Widgetidget)

    def set_layout_editorR(self):  # noqa: N802
        """Set Layout to Editor right, PDF left."""
        # Horizontal orientation
        self.splitter.setOrientation(QtCore.Qt.Horizontal)

        # Reorder widgets
        self.splitter.addWidget(self.Widgetidget_31)
        self.splitter.addWidget(self.Widgetidget)

    def set_layout_editorL(self):  # noqa: N802
        """Set Layout to Editor left, PDF right."""
        # Horizontal orientation
        self.splitter.setOrientation(QtCore.Qt.Horizontal)

        # Reorder widgets
        self.splitter.addWidget(self.Widgetidget)
        self.splitter.addWidget(self.Widgetidget_31)

    def use_default_layout(self):
        """Adjust the layout set in the config."""
        match config.get("Layout", "default_layout"):
            case "typewriter":
                self.actions.layout_typewriter.trigger()
            case "editor_right":
                self.actions.layout_editorR.trigger()
            case "editor_left":
                self.actions.layout_editorL.trigger()
            case _:
                self.actions.layout_typewriter.trigger()

        self.actions.show_fs_explorer.setChecked(config.get("Layout", "show_fs_explorer", typ="bool"))
        self.actions.show_compiler_options.setChecked(config.get("Layout", "show_compiler_options", typ="bool"))
        self.actions.show_compiler_output.setChecked(config.get("Layout", "show_compiler_output", typ="bool"))

    def set_fs_explorer_visibility(self, visibility):
        """Set the visibility of the fs explplorer."""
        self.FSdock.setVisible(visibility)

    def set_compiler_options_visibility(self, visibility):
        """Set the visibility of the fs explplorer."""
        self.CompilerOptionsdock.setVisible(visibility)

    def set_compiler_output_visibility(self, visibility):
        """Set the visibility of the fs explplorer."""
        self.CompilerOutputdock.setVisible(visibility)

    def open_config(self):
        """Open config file."""
        config.write()
        util.open_with_external_program(config.writepath)

    def closeEvent(self, event):  # noqa: N802
        """Handle close event."""
        self.CompilerConnector.stop()
        self.save_session()
        s = self.editor.tryclose()
        if s:
            event.accept()
        else:
            event.ignore()

    def prepare_compilation(self):
        """Set appropriate input and output files before compiling."""
        if state.main_file.Value is None:  # noqa: SIM108
            main = self.editor.TabWidget.currentWidget().path
        else:
            main = state.main_file.Value

        if main:
            self.CompilerConnector.set_fin(main)
            self.CompilerConnector.set_fout(util.pdf_path(main))
            self.PDFWidget.open(util.pdf_path(main))

    def load_session(self):
        """Load the last session."""
        logger.info("Loading last session.")
        session = util.read_session_file()
        if session:
            (working_directory, files) = session
            state.working_directory.Value = working_directory
            for f in files:
                self.editor.open_file(f)

    def save_session(self):
        """Save the current session."""
        logger.debug("Saving last session.")
        util.write_session_file(state.working_directory.Value, self.editor.openfiles_list())

    def check_typst_availability(self):
        """Check if typst is available."""
        if not util.typst_available():
            logger.warning("The typst executable was not found.")

            msg_box = QtWidgets.QMessageBox()
            msg_box.setIcon(QtWidgets.QMessageBox.Critical)
            msg_box.setWindowTitle("Typst not found")
            msg_box.setText(
                "The typst executable could not be found found:\nTypstwriter will not be able to compile documents."
            )
            msg_box.setInformativeText(
                "Please install typst, make sure it was added to the 'PATH' environment variable \
and check the typstwriter configuration file."
            )
            msg_box.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Ignore | QtWidgets.QMessageBox.StandardButton.Abort)
            msg_box.setDefaultButton(QtWidgets.QMessageBox.StandardButton.Ignore)

            match msg_box.exec():
                case QtWidgets.QMessageBox.StandardButton.Ignore:
                    pass
                case QtWidgets.QMessageBox.StandardButton.Abort:
                    self.close()



import argparse

import typstwriter


class APA(argparse.ArgumentParser):
    """
    A customized subclass of argparse.ArgumentParser.

    Displays the description before the usage in the help output, capitalizes the section titles and adds a prolog section.
    """

    def __init__(self, *args, prolog=None, **kwargs):
        """Patch positionals and optionals display names."""
        argparse.ArgumentParser.__init__(self, *args, **kwargs)
        self.prolog = prolog
        self._positionals.title = "Arguments"
        self._optionals.title = "Options"

    def format_usage(self):
        """Format the usage text."""
        formatter = self._get_formatter()
        formatter.add_usage(self.usage, self._actions, self._mutually_exclusive_groups, "Usage: ")
        return formatter.format_help()

    def format_help(self):
        """Format the help text."""
        formatter = self._get_formatter()

        # description
        formatter.add_text(self.description)

        # prolog
        formatter.add_text(self.prolog)

        # usage
        formatter.add_usage(self.usage, self._actions, self._mutually_exclusive_groups, "Usage: ")

        # positionals, optionals and user-defined groups
        for action_group in self._action_groups:
            formatter.start_section(action_group.title)
            formatter.add_text(action_group.description)
            formatter.add_arguments(action_group._group_actions)
            formatter.end_section()

        # epilog
        formatter.add_text(self.epilog)

        # format help message
        return formatter.format_help()


def parse_args():
    """Parse CLI arguments."""
    parser = APA(description=typstwriter.__doc__, add_help=False)

    parser.add_argument("files", nargs="*", help="The list of files to open.")

    parser.add_argument(
        "-h",
        "--help",
        action="help",
        help="Show this help message and exit.",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"Typstwriter {typstwriter.__version__}",
        help="Show the Typstwriter version and exit.",
    )
    parser.add_argument(
        "--working-directory",
        action="store",
        type=str,
        help="Set the working directory.",
    )

    return parser.parse_args()


Args = parse_args()



import sys
import signal
import os


os.environ["QT_API"] = "pyside6"
import qtpy  # noqa: E402 RUF100


def main():
    """Run Typstwriter."""
    # Initialise logging
    from typstwriter import logging

    logging.setup_logger(os.environ.get("LOGLEVEL"))
    logger = logging.getLogger(__name__)
    logger.debug("Logging initialized")

    # Parse Arguments
    logger.debug("Parse Arguments")
    from typstwriter import arguments  # noqa: F401

    # Start Typstwriter
    logger.info("Typstwriter started")

    # Initialise Config
    logger.info("Reading Config")
    from typstwriter import configuration  # noqa: F401

    # Initialise State
    logger.info("Initialising State")
    from typstwriter import globalstate  # noqa: F401

    # With logging, config and state set up, import the main GUI
    from typstwriter import mainwindow

    # Make sure the application can receive SIGINT
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    # Initialize the application
    logger.info("Initialize Application")
    app = qtpy.QtWidgets.QApplication(sys.argv)
    main = mainwindow.MainWindow()
    main.show()

    # Run he application
    logger.info("Run Application")
    app.exec_()


if __name__ == "__main__":
    main()



from qtpy import QtCore
from qtpy import QtGui

import logging

from typstwriter import util


class TestFileIconProvider:
    """Test util.FileIconProvider."""

    def test_dottyp(self, qtbot):
        """Test if the .typ icon is available."""
        finfo = QtCore.QFileInfo(QtCore.QDir("some/path/"), "file.typ")
        fip = util.FileIconProvider()
        icon = fip.icon(finfo)
        assert isinstance(icon, QtGui.QIcon)

    def test_dotother(self, qtbot):
        """Test if the .typ icon is available."""
        finfo = QtCore.QFileInfo(QtCore.QDir("some/path/"), "file.other")
        fip = util.FileIconProvider()
        icon = fip.icon(finfo)
        assert isinstance(icon, QtGui.QIcon)


class TestRecentFilesModel:
    """Test util.RecentFilesModel."""

    file_list = ["a.txt", "b.typ", "c.py"]  # noqa RUF012

    def test_data_data(self):
        """Make sure correct data is returned."""
        model = util.RecentFilesModel(self.file_list)
        for i, f in enumerate(self.file_list):
            assert model.data(model.createIndex(i, 0, None), QtCore.Qt.DisplayRole) == f

    def test_data_icon(self):
        """Make sure an icon is returned."""
        model = util.RecentFilesModel(self.file_list)
        for i, _ in enumerate(self.file_list):
            assert isinstance(model.data(model.createIndex(i, 0, None), QtCore.Qt.DecorationRole), QtGui.QIcon)

    def test_data_other(self):
        """Make sure an None is returned for other role."""
        model = util.RecentFilesModel(self.file_list)
        for i, _ in enumerate(self.file_list):
            assert model.data(model.createIndex(i, 0, None), QtCore.Qt.EditRole) is None

    def test_rowCount(self):
        """Make sure the row_count is correct."""
        model = util.RecentFilesModel(self.file_list)
        assert model.rowCount(0) == len(self.file_list)

    def test_header(self):
        """Make sure the header is a string."""
        model = util.RecentFilesModel([])
        header = model.headerData(0, QtCore.Qt.Vertical)
        assert isinstance(header, str)


class TestRecentFiles:
    """Test util.RecentFiles."""

    def test_successful_read(self, tmp_path):
        """Test if reading a valid file is successful."""
        p = tmp_path / "recent_files"
        p.write_text("a.typ\nb.txt\nfolder/c.typ")
        r = util.RecentFiles(p)
        assert r.list() == ["a.typ", "b.txt", "folder/c.typ"]

    def test_successful_write(self, tmp_path):
        """Test if writing a valid file is successful."""
        p = tmp_path / "recent_files"
        p.write_text("")
        r = util.RecentFiles(p)
        r.append("a.txt")
        r.append("b.txt")
        r.append("folder/c.txt")
        r.write()
        assert p.read_text() == "folder/c.txt\nb.txt\na.txt\n"

    def test_failed_read(self, tmp_path, caplog):
        """Test if nonexistent file fails."""
        caplog.set_level(logging.INFO)
        p = tmp_path / "inexistent_file"
        util.RecentFiles(p)
        assert "Could not read" in caplog.text

    def test_failed_write(self, tmp_path, caplog):
        """Test if writing to a write protected file fails."""
        caplog.set_level(logging.INFO)
        p = tmp_path / "recent_files"
        p.write_text("")
        r = util.RecentFiles(p)
        p.chmod(0o000)
        r.write()
        assert "Could not write" in caplog.text

    def test_manipulation(self, tmp_path):
        """Test manipulationn of the recent files list."""
        # Create temporary recent files file
        p = tmp_path / "recent_files"
        p.write_text("")

        # Actual tests
        r = util.RecentFiles(p)
        r.append("a.txt")
        assert r.list() == ["a.txt"]
        r.append("b.txt")
        assert r.list() == ["b.txt", "a.txt"]
        r.append("folder/c.txt")
        assert r.list() == ["folder/c.txt", "b.txt", "a.txt"]
        r.append("b.txt")
        assert r.list() == ["b.txt", "folder/c.txt", "a.txt"]
        r.clear()
        assert r.list() == []


def test_pdf_path():
    """Test util.pdf_path()."""
    assert util.pdf_path("a/b/c/d.typ") == "a/b/c/d.pdf"
    assert util.pdf_path("a/b/c/d.txt") == "a/b/c/d.pdf"
    assert util.pdf_path("a/b/c/d") == "a/b/c/d.pdf"
    assert util.pdf_path("a/b/c/.d.typ") == "a/b/c/.d.pdf"
    assert util.pdf_path("d.typ") == "d.pdf"
    assert util.pdf_path(".d.typ") == ".d.pdf"


class TestTogglingAction:
    """Test util.TestTogglingAction."""

    def test_text_toggle(self):
        """Make sure the text toggles."""
        action = util.TogglingAction()
        action.setText("Off", state=QtGui.QIcon.State.Off)
        action.setText("On", state=QtGui.QIcon.State.On)

        action.setChecked(True)
        assert action.text() == "On"

        action.toggle()
        assert action.text() == "Off"

        action.toggle()
        assert action.text() == "On"

    def test_triggered_signal(self, qtbot):
        """Make sure the activated and deactivated signals are emitted."""
        action = util.TogglingAction()

        action.setChecked(True)

        with qtbot.waitSignal(action.deactivated):
            action.trigger()

        with qtbot.waitSignal(action.activated):
            action.trigger()


def test_typst_available(monkeypatch):
    """Test util.typst_available."""
    monkeypatch.setattr(QtCore.QProcess, "start", lambda *args: None)
    monkeypatch.setattr(QtCore.QProcess, "exitCode", lambda *args: 2)
    monkeypatch.setattr(QtCore.QProcess, "exitStatus", lambda *args: QtCore.QProcess.ExitStatus.NormalExit)

    # Typst available
    monkeypatch.setattr(QtCore.QProcess, "readAll", lambda *args: b"The Typst compiler")
    assert util.typst_available()

    # Typst not available
    monkeypatch.setattr(QtCore.QProcess, "readAll", lambda *args: b"Some other program")
    assert not util.typst_available()


def test_qstring_length():
    """Test util.qstring_length."""
    assert util.qstring_length("asdf") == 4  # noqa: PLR2004
    assert util.qstring_length("Some\nText") == 9  # noqa: PLR2004
    assert util.qstring_length("👍") == 2  # noqa: PLR2004
    assert util.qstring_length('"Happy Birthday 🎉"') == 19  # noqa: PLR2004



from typstwriter import logging
from re import match


def test_Message():
    """Test logging.Message."""
    assert str(logging.Message("A string: {}", ["Test"])) == "A string: Test"
    assert str(logging.Message("A debug string: {!r}", ["Test"])) == "A debug string: 'Test'"
    assert str(logging.Message("{} + {} = {}", [2, 2, 4])) == "2 + 2 = 4"


def test_StyleAdapter(caplog):
    """Test logging.StyleAdapter."""
    sa = logging.getLogger("test")

    sa.log(100, "A test")
    assert "A test" in caplog.text
    caplog.clear()

    sa.log(100, "A test with arguments: {}, {}", [1, 2, 3], "0")
    assert "A test with arguments: [1, 2, 3], 0" in caplog.text
    caplog.clear()

    sa.log(0, "A test")
    assert not caplog.text
    caplog.clear()


# This function does not use the caplog fixture directly because caplog injects
# its own handler in the logging and overrides the custom formatting.
# Instead the whole output is captured.
def test_setup_logger(capsys):
    """Test logging.getLogger."""
    logging.setup_logger("WARNING")
    logger = logging.getLogger("test")

    logger.critical("A test")
    (out, err) = capsys.readouterr()
    assert match(r"CRITICAL\s+\S+ \S+\s+test_logging.py\s+test_setup_logger\s+line \d+\s+: A test", err)

    logger.info("A test")
    (out, err) = capsys.readouterr()
    assert not out
    assert not err

    logging.setup_logger(level="Invalid")
    logger = logging.getLogger("test")

    logger.info("A test")
    (out, err) = capsys.readouterr()
    assert match(r"INFO\s+\S+ \S+\s+test_logging.py\s+test_setup_logger\s+line \d+\s+: A test", err)



from qtpy import QtWidgets

from typstwriter import compiler_tools

from typstwriter import globalstate


class TestCompilerOptions:
    """Test compiler_tools.CompilerOptions."""

    def test_main_changed(self, qtbot, tmp_path):
        """Check the behaviour when main changes."""
        pth = tmp_path / "some_file.typ"
        compileroptions = compiler_tools.CompilerOptions()
        compileroptions.main_changed(str(pth))
        assert compileroptions.line_edit_main.text() == str(pth)

    def test_main_path_edited(self, qtbot, tmp_path):
        """Check the behaviour when the main_path is edited."""
        pth = tmp_path / "some_file.typ"
        compileroptions = compiler_tools.CompilerOptions()

        # invalid path
        compileroptions.line_edit_main.setText(str(pth))
        compileroptions.line_edit_main.editingFinished.emit()
        assert compileroptions.line_edit_main.text() != str(pth)
        assert globalstate.State.main_file.Value != str(pth)

        # valid path
        pth.touch()
        compileroptions.line_edit_main.setText(str(pth))
        compileroptions.line_edit_main.editingFinished.emit()
        assert compileroptions.line_edit_main.text() == str(pth)
        assert globalstate.State.main_file.Value == str(pth)

    def test_file_dialog(self, qtbot, tmp_path, monkeypatch):
        """Test setting the main file via file dialog."""
        pth = tmp_path / "some_file.typ"
        pth.touch()
        compileroptions = compiler_tools.CompilerOptions()
        monkeypatch.setattr(QtWidgets.QFileDialog, "getOpenFileName", lambda *args: (str(pth), ""))

        compileroptions.open_file_dialog()
        assert compileroptions.line_edit_main.text() == str(pth)
        assert globalstate.State.main_file.Value == str(pth)

    def test_mode_changed(self, qtbot, tmp_path):
        """Test canging mode."""
        pth = tmp_path / "some_file.typ"
        pth.touch()
        compileroptions = compiler_tools.CompilerOptions()
        compileroptions.combo_box_mode.setCurrentIndex(0)
        assert globalstate.State.compiler_mode.Value == compileroptions.combo_box_mode.itemData(0)
        compileroptions.combo_box_mode.setCurrentIndex(1)
        assert globalstate.State.compiler_mode.Value == compileroptions.combo_box_mode.itemData(1)


class TestCompilerOutput:
    """Test compiler_tools.CompilerOutput."""

    def test_clear_button(self, qtbot):
        """Make sure clearing the display works."""
        compileroutput = compiler_tools.CompilerOutput()
        compileroutput.OutputDisplay.insertPlainText("Test")
        assert compileroutput.OutputDisplay.toPlainText() == "Test"
        compileroutput.ClearDisplay.click()
        assert compileroutput.OutputDisplay.toPlainText() == ""

    def test_insert_insert_block(self, qtbot):
        """Test compiler_tools.insert_block."""
        compileroutput = compiler_tools.CompilerOutput()

        compileroutput.insert_block(text="Test")
        assert compileroutput.OutputDisplay.toPlainText() == "Test"
        assert "<hr />" not in compileroutput.OutputDisplay.toHtml()
        compileroutput.insert_block(text="Test")
        assert "<hr />" in compileroutput.OutputDisplay.toHtml()
        compileroutput.OutputDisplay.clear()

    def test_append_to_block(self, qtbot):
        """Test compiler_tools.append_to_block."""
        compileroutput = compiler_tools.CompilerOutput()

        compileroutput.append_to_block(text="Test\n")
        compileroutput.append_to_block(text="Test")
        assert compileroutput.OutputDisplay.toPlainText() == "Test\nTest"
        assert "<hr />" not in compileroutput.OutputDisplay.toHtml()



from qtpy import QtPdfWidgets
from qtpy import QtPdf

import pytest
from unittest.mock import Mock
import logging
from fpdf import FPDF


from typstwriter import pdf_viewer
from typstwriter import util


@pytest.fixture()
def tmp_pdf(tmp_path, pdf_pages):
    """Create an pdf document with "pdf_pages" empty pages and return its location."""
    n_max = int(1e4)

    for n in range(n_max):
        path = tmp_path / f"tmp_pdf_{n:04d}.pdf"
        if not path.exists():
            break

    pdf = FPDF()
    pdf.set_font("Times", "", 12)
    for _ in range(pdf_pages):
        pdf.add_page()
    pdf.output(path, "F")

    with path.open() as f:
        pdf.write(f)

    return path


class TestZoomSelector:
    """Test pdf_viewer.ZoomSelector."""

    @pytest.mark.parametrize("zoom", [1.0, 0.5, 4.0])
    def test_numeric_user_input(self, qtbot, zoom):
        """Make sure the zoom factor can be set."""
        zoomselector = pdf_viewer.ZoomSelector(None)
        zoomselector.lineEdit().setText("")

        with qtbot.waitSignals(
            [(zoomselector.zoom_mode_changed, "zoom_mode_changed"), (zoomselector.zoom_factor_changed, "zoom_factor_changed")]
        ) as blocker:
            zoomselector.lineEdit().setText(f"{int(zoom * 100)}%")

        args = [s.args for s in blocker.all_signals_and_args]
        assert (zoom,) in args
        assert (QtPdfWidgets.QPdfView.ZoomMode.Custom,) in args

    @pytest.mark.parametrize(
        ("text", "mode"),
        [
            ("Fit Page Width", QtPdfWidgets.QPdfView.ZoomMode.FitToWidth),
            ("Fit Page Height", QtPdfWidgets.QPdfView.ZoomMode.FitInView),
        ],
    )
    def test_mode_user_input(self, qtbot, text, mode):
        """Make sure the zoom factor can be set."""
        zoomselector = pdf_viewer.ZoomSelector(None)
        zoomselector.lineEdit().setText("")

        with qtbot.waitSignal((zoomselector.zoom_mode_changed, "zoom_mode_changed")) as blocker:
            zoomselector.lineEdit().setText(text)
            print(f"{text!r}")
            print(f"{zoomselector.lineEdit().text()!r}")

        assert mode in blocker.args

    @pytest.mark.parametrize("zoom", [0.1, 0.5, 1.0, 1.5, 2])
    def test_set_zoom_factor(self, qtbot, zoom):
        """Make sure the zoom factor can be set."""
        zoomselector = pdf_viewer.ZoomSelector(None)

        with qtbot.waitSignals(
            [(zoomselector.zoom_mode_changed, "zoom_mode_changed"), (zoomselector.zoom_factor_changed, "zoom_factor_changed")]
        ) as blocker:
            zoomselector.set_zoom_factor(zoom)

        args = [s.args for s in blocker.all_signals_and_args]
        assert (zoom,) in args
        assert (QtPdfWidgets.QPdfView.ZoomMode.Custom,) in args

    def test_reset(self, qtbot):
        """Test reset functionality."""
        zoomselector = pdf_viewer.ZoomSelector(None)

        with qtbot.waitSignals(
            [(zoomselector.zoom_mode_changed, "zoom_mode_changed"), (zoomselector.zoom_factor_changed, "zoom_factor_changed")]
        ) as blocker:
            zoomselector.reset()

        args = [s.args for s in blocker.all_signals_and_args]
        assert (1.0,) in args
        assert (QtPdfWidgets.QPdfView.ZoomMode.Custom,) in args


class TestPDFViewer:
    """Test pdf_viewer.PDFViewer."""

    @pytest.mark.parametrize("pdf_pages", [1, 2])
    def test_open_file_fail(self, qtbot, tmp_path, tmp_pdf, caplog):
        """Test opening a path that does not exist."""
        caplog.set_level(logging.INFO)
        pdf_v = pdf_viewer.PDFViewer()

        # Failure
        document = tmp_path / "document.pdf"
        pdf_v.open(str(document))
        assert pdf_v.m_document.status() == QtPdf.QPdfDocument.Status.Null
        assert "not a valid file" in caplog.text

        caplog.clear()

        # Success
        pdf_v.open(str(tmp_pdf))
        assert pdf_v.m_document.status() == QtPdf.QPdfDocument.Status.Ready
        assert "not a valid file" not in caplog.text

        # TODO: Test opening non pdf file

    @pytest.mark.parametrize("pdf_pages", [1, 2])
    def test_reload(self, qtbot, caplog, tmp_pdf):
        """Test reloading a pdf file."""
        caplog.set_level(logging.DEBUG)
        pdf_v = pdf_viewer.PDFViewer()

        pdf_v.open(str(tmp_pdf))

        sb = pdf_v.pdfView.verticalScrollBar()
        sb.setValue((sb.maximum() - sb.minimum()) / 2)
        pos = sb.value()

        pdf_v.reload()
        assert pdf_v.m_document.status() == QtPdf.QPdfDocument.Status.Ready
        assert "Reloaded from disk" in caplog.text
        assert pos == sb.value()

    @pytest.mark.parametrize("zoom", [0.6, 0.9, 1.2, 2.4])
    @pytest.mark.parametrize("pdf_pages", [1, 2])
    def test_zoom(self, qtbot, caplog, tmp_pdf, zoom):
        """Test reloading a pdf file."""
        caplog.set_level(logging.DEBUG)
        pdf_v = pdf_viewer.PDFViewer()

        pdf_v.open(str(tmp_pdf))

        pdf_v.m_zoomSelector.setCurrentText(f"{int(zoom * 100)}%")
        assert pdf_v.m_zoomSelector.currentText() == f"{int(zoom * 100)}%"
        assert pdf_v.pdfView.zoomFactor() == zoom

        pdf_v.m_zoomSelector.setCurrentText(f"{int(zoom * 100)}%")
        pdf_v.actionZoom_Out.trigger()
        assert pdf_v.m_zoomSelector.currentText() == f"{int(zoom * 100 / 1.2)}%"
        assert pdf_v.pdfView.zoomFactor() == zoom / 1.2

        pdf_v.m_zoomSelector.setCurrentText(f"{int(zoom * 100)}%")
        pdf_v.actionZoom_In.trigger()
        assert pdf_v.m_zoomSelector.currentText() == f"{int(zoom * 100 * 1.2)}%"
        assert pdf_v.pdfView.zoomFactor() == zoom * 1.2

        # TODO: test fit width
        # TODO; test fit height

    @pytest.mark.parametrize("pdf_pages", [1, 4, 10])
    def test_page_movement(self, qtbot, caplog, tmp_pdf, pdf_pages):
        """Test reloading a pdf file."""
        caplog.set_level(logging.DEBUG)
        pdf_v = pdf_viewer.PDFViewer()

        pdf_v.open(str(tmp_pdf))

        assert pdf_v.m_maxPage.text() == f" of {pdf_pages}"

        assert pdf_v.m_pageSelector.text() == str(min(1, pdf_pages))
        assert pdf_v.pdfView.pageNavigator().currentPage() + 1 == min(1, pdf_pages)

        pdf_v.actionPrevious_Page.trigger()
        assert pdf_v.m_pageSelector.text() == str(min(1, pdf_pages))
        assert pdf_v.pdfView.pageNavigator().currentPage() + 1 == min(1, pdf_pages)

        pdf_v.actionNext_Page.trigger()
        assert pdf_v.m_pageSelector.text() == str(min(2, pdf_pages))
        assert pdf_v.pdfView.pageNavigator().currentPage() + 1 == min(2, pdf_pages)

        pdf_v.m_pageSelector.setValue(4)
        assert pdf_v.m_pageSelector.text() == str(min(4, pdf_pages))
        assert pdf_v.pdfView.pageNavigator().currentPage() + 1 == min(4, pdf_pages)

        pdf_v.actionPrevious_Page.trigger()
        assert pdf_v.m_pageSelector.text() == str(min(3, pdf_pages))
        assert pdf_v.pdfView.pageNavigator().currentPage() + 1 == min(3, pdf_pages)

        # TODO: Test scrolling

    @pytest.mark.parametrize("pdf_pages", [1, 2])
    def test_open_in_external_viewer(self, qtbot, caplog, tmp_pdf, monkeypatch):
        """Test opening a pdf file with an external viewer."""
        external_program = Mock()
        monkeypatch.setattr(util, "open_with_external_program", external_program)

        caplog.set_level(logging.INFO)
        pdf_v = pdf_viewer.PDFViewer()

        pdf_v.actionOpen_External.trigger()
        assert "Attempted to open external PDF viewer but no file is loaded." in caplog.text
        external_program.assert_not_called()

        pdf_v.open(str(tmp_pdf))

        pdf_v.actionOpen_External.trigger()
        external_program.assert_called_once()



from qtpy import QtCore

import os.path

from typstwriter import enums

from typstwriter import logging
from typstwriter import configuration
from typstwriter import arguments

logger = logging.getLogger(__name__)
config = configuration.Config
args = arguments.Args


class SingleState(QtCore.QObject):
    """Stores a single state in 'Value' and emits 'Signal(value)' whenever 'Value' is reassigned."""

    Signal = QtCore.Signal(object)
    Value = None

    def __init__(self, value=None):
        """Initialize and assign value, if present."""
        QtCore.QObject.__init__(self)
        self.Value = value

    def __setattr__(self, name, value):
        """Emit Signal upon reassignment of Value. Reject assigning any other attribute."""
        if name == "Value":
            super().__setattr__(name, value)
            self.Signal.emit(value)
        elif name == "Signal":
            super().__setattr__(name, value)
        else:
            raise AttributeError("The only valid attributes are 'Value' and 'Signal'.")


class GlobalState:
    """Stores and initializes a collection of SingleStates."""

    def __init__(self):
        """Initiate members according to config and args."""
        wd = os.path.expanduser("~/")
        if args.working_directory:
            if os.path.exists(os.path.expanduser(args.working_directory)):
                wd = os.path.expanduser(args.working_directory)
            else:
                logger.warning("The argument working-directory={!r} is not a valid path.", args.working_directory)
        elif os.path.exists(os.path.expanduser(config.get("General", "working_directory"))):
            wd = os.path.expanduser(config.get("General", "working_directory"))

        self.working_directory = SingleState(os.path.abspath(wd))
        self.compiler_running = SingleState(False)
        self.main_file = SingleState(None)
        self.compiler_mode = SingleState(enums.compiler_mode[config.get("Compiler", "mode")])


# Instantiate the State singleton
State = GlobalState()



from qtpy import QtCore
from qtpy import QtWidgets

from typstwriter import logging
from typstwriter import configuration
from typstwriter import globalstate

logger = logging.getLogger(__name__)
config = configuration.Config
state = globalstate.State


class ToolBar(QtWidgets.QToolBar):
    """Main Window Tool Bar."""

    def __init__(self, actions):
        """Populate Tool Bar."""
        QtWidgets.QToolBar.__init__(self)

        self.setMovable(False)
        self.setFloatable(False)
        self.setToolButtonStyle(QtCore.Qt.ToolButtonStyle.ToolButtonTextBesideIcon)

        self.addAction(actions.new_File)
        self.addAction(actions.open_File)
        self.addSeparator()
        self.addAction(actions.save)
        self.addAction(actions.save_as)
        self.addSeparator()
        self.addAction(actions.run)



from qtpy import QtGui
from qtpy import QtCore
from qtpy import QtWidgets

import os
import shutil

from typstwriter import util

from typstwriter import logging
from typstwriter import configuration
from typstwriter import globalstate

logger = logging.getLogger(__name__)
config = configuration.Config
state = globalstate.State


class FSContextMenu(QtWidgets.QMenu):
    """
    Base context menu for the fs Explorer.

    This class contains all actions and handlers that may be needed in a FS context menu.
    It does not actually construct a menu however, this is done in a subclass which can
    use a selection of the actions declared in this clas to populate the menu.
    """

    def __init__(self, parent):
        """Declare all actions."""
        QtWidgets.QMenu.__init__(self, parent)
        self.context_path = None

        self.action_open = QtWidgets.QAction("Open", triggered=self.handle_open)
        self.action_open_external = QtWidgets.QAction("Open with external program", triggered=self.handle_open_external)
        self.action_new_file = QtWidgets.QAction("New file", triggered=self.handle_new_file)
        self.action_new_folder = QtWidgets.QAction("New folder", triggered=self.handle_new_folder)
        self.action_rename = QtWidgets.QAction("Rename", triggered=self.handle_rename)
        self.action_delete = QtWidgets.QAction("Delete", triggered=self.handle_delete)
        self.setRoot = QtWidgets.QAction("Set as Working Directory", triggered=self.handle_set_root)
        self.action_copy_path = QtWidgets.QAction("Copy path", triggered=self.handle_copy_path)
        self.action_copy = QtWidgets.QAction("Copy", triggered=self.handle_copy)
        self.action_paste = QtWidgets.QAction("Paste", triggered=self.handle_paste)
        self.action_main_file = QtWidgets.QAction("Set as main file", triggered=self.handle_set_as_main_file)

    def handle_open(self):
        """Trigger opening a file."""
        self.parent().open_file.emit(self.context_path)

    def handle_open_external(self):
        """Trigger opening a file."""
        util.open_with_external_program(self.context_path)

    def handle_new_file(self):
        """Trigger creating a new file."""
        self.parent().new_file_in_dir(self.context_path)

    def handle_new_folder(self):
        """Trigger creating a new folder."""
        self.parent().new_folder_in_dir(self.context_path)

    def handle_rename(self):
        """Trigger renaming folder."""
        self.parent().rename_from(self.context_path)

    def handle_delete(self):
        """Trigger deleteing file or folder."""
        self.parent().delete(self.context_path)

    def handle_set_root(self):
        """Trigger setting the current folder as working directory."""
        self.parent().set_root(self.context_path)

    def handle_copy_path(self):
        """Trigger copying path to the clipboard."""
        QtGui.QGuiApplication.clipboard().setText(self.context_path)

    def handle_copy(self):
        """Trigger copying file or folder to the clipboard."""
        self.parent().copy_to_clipboard(self.context_path)

    def handle_paste(self):
        """Trigger pasting from the clipboard."""
        self.parent().paste_from_clipboard(self.context_path)

    def handle_set_as_main_file(self):
        """Trigger setting the current file as main file."""
        state.main_file.Value = self.context_path


class NoItemContextMenu(FSContextMenu):
    """Context menu when not clicking on any item."""

    def __init__(self, parent):
        """Assemble Menu."""
        FSContextMenu.__init__(self, parent)

        self.addAction(self.action_new_file)
        self.addAction(self.action_new_folder)
        self.addAction(self.action_paste)


class FileContextMenu(FSContextMenu):
    """Context menu when clicking on a file."""

    def __init__(self, parent):
        """Assemble Menu."""
        FSContextMenu.__init__(self, parent)

        self.addAction(self.action_open)
        self.addAction(self.action_open_external)
        self.addAction(self.action_rename)
        self.addAction(self.action_delete)
        self.addAction(self.action_copy)
        self.addAction(self.action_copy_path)
        self.addAction(self.action_main_file)


class FolderContextMenu(FSContextMenu):
    """Context menu when clicking on a folder."""

    def __init__(self, parent):
        """Assemble Menu."""
        FSContextMenu.__init__(self, parent)

        self.addAction(self.action_new_file)
        self.addAction(self.action_new_folder)
        self.addAction(self.action_copy)
        self.addAction(self.action_copy_path)
        self.addAction(self.action_paste)
        self.addAction(self.action_rename)
        self.addAction(self.action_delete)
        self.addAction(self.setRoot)


class SelectionContextMenu(FSContextMenu):
    """Context menu when clicking on a selection."""

    def __init__(self, parent):
        """Assemble Menu."""
        FSContextMenu.__init__(self, parent)

        self.addAction(self.action_delete)
        self.addAction(self.action_copy)


class FSExplorer(QtWidgets.QWidget):
    """A filesystem explorer widget."""

    # directoryChanged = QtCore.Signal(str)
    open_file = QtCore.Signal(str)

    def __init__(self):
        """Populate widget and set initial state."""
        QtWidgets.QWidget.__init__(self)

        # Populate Widget
        self.Layout = QtWidgets.QVBoxLayout(self)
        self.Layout.setContentsMargins(4, 4, 4, 4)
        self.Layout.setSpacing(2)

        self.pathBar = QtWidgets.QLineEdit(self)
        self.pathBar.setText("")
        self.pathBar.editingFinished.connect(self.line_edited)

        self.folderAction = QtWidgets.QAction(
            QtGui.QIcon.fromTheme(QtGui.QIcon.FolderOpen, QtGui.QIcon(util.icon_path("folder.svg"))), "open"
        )
        self.folderAction.triggered.connect(self.open_directory_dialog)
        self.pathBar.addAction(self.folderAction, QtWidgets.QLineEdit.LeadingPosition)

        self.parentAction = QtWidgets.QAction(
            QtGui.QIcon.fromTheme(QtGui.QIcon.GoUp, QtGui.QIcon(util.icon_path("parent_dir.svg"))), "up"
        )
        self.parentAction.triggered.connect(self.goto_parent_directory)
        self.pathBar.addAction(self.parentAction, QtWidgets.QLineEdit.TrailingPosition)

        self.tree_view = QtWidgets.QTreeView()
        self.filesystem_model = QtWidgets.QFileSystemModel(self.tree_view)
        self.filesystem_model.setIconProvider(util.FileIconProvider())
        self.tree_view.setModel(self.filesystem_model)
        self.tree_view.setRootIsDecorated(True)
        self.tree_view.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.tree_view.hideColumn(1)
        self.tree_view.hideColumn(2)
        self.tree_view.header().setSectionsMovable(False)
        self.tree_view.header().setStretchLastSection(False)
        self.tree_view.header().setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeMode.Stretch)
        self.tree_view.header().setSectionResizeMode(3, QtWidgets.QHeaderView.ResizeMode.ResizeToContents)
        self.tree_view.doubleClicked.connect(self.doubleclicked)

        # A second filesystem model is needed for the completer because only directories should be shown as completions
        # while both files and directories should be shown in the tree view.
        # This duplication is somewhat unsatisfying but it was decently simple to implement and the overhead is acceptable.
        self.completer_filesystem_model = QtWidgets.QFileSystemModel()
        self.completer_filesystem_model.setFilter(QtCore.QDir.AllDirs | QtCore.QDir.NoDotAndDotDot)
        self.completer = QtWidgets.QCompleter()
        self.completer.setModel(self.completer_filesystem_model)
        self.pathBar.setCompleter(self.completer)

        self.Layout.addWidget(self.pathBar)
        self.Layout.addWidget(self.tree_view)

        # Add Context Menu
        self.tree_view.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.tree_view.customContextMenuRequested.connect(self.rightclicked)
        self.NoItemContextMenu = NoItemContextMenu(self)
        self.FileContextMenu = FileContextMenu(self)
        self.FolderContextMenu = FolderContextMenu(self)
        self.SelectionContextMenu = SelectionContextMenu(self)

        # Set initial state
        self.set_root(os.path.normpath(state.working_directory.Value))

    def open_directory_dialog(self):
        """Open a dialog to select root path and open said path."""
        directory = QtWidgets.QFileDialog.getExistingDirectory(
            self,
            "Select Working Directory",
            self.pathBar.text(),
            QtWidgets.QFileDialog.ShowDirsOnly | QtWidgets.QFileDialog.DontResolveSymlinks,
        )

        if os.path.isdir(directory):
            self.set_root(directory)

    def set_root(self, root):
        """Set the root directory of the FSExplorer."""
        state.working_directory.Value = root
        self.root_changed(root)

    @QtCore.Slot(str)
    def root_changed(self, root):
        """Update root directory."""
        self.root = root
        self.filesystem_model.setRootPath(root)
        self.tree_view.setRootIndex(self.filesystem_model.index(root))
        self.completer_filesystem_model.setRootPath(root)
        self.pathBar.setText(root)
        logger.debug("Change Working Directory to {!r}.", root)
        # self.directoryChanged.emit(root)

    @QtCore.Slot()
    def line_edited(self):
        """Handle LineEdits editingFinished signal. Change root directory if path is valid and resets otherwise."""
        root = os.path.normpath(os.path.expanduser(self.pathBar.text()))
        if os.path.isdir(root):
            self.set_root(root)
        else:
            self.pathBar.setText(self.root)
            logger.info("Attempted to change Working Directory but {!r} is not a valid directory.", root)

    @QtCore.Slot(QtCore.QModelIndex)
    def doubleclicked(self, index):
        """Handle doubleclick on tree-view entry. Open file or changes root directory."""
        path = self.filesystem_model.filePath(index)
        if os.path.isfile(path):
            self.open_file.emit(path)
        if os.path.isdir(path):
            self.set_root(path)

    @QtCore.Slot()
    def goto_parent_directory(self):
        """Set the parent of the current directory as root."""
        path = os.path.dirname(self.root)
        self.set_root(path)

    def rightclicked(self, event):
        """Handle right click."""
        index = self.tree_view.indexAt(event)
        if not index.isValid():
            self.open_no_item_menu()
            return
        else:
            selected_indices = self.tree_view.selectionModel().selectedRows()
            if index in selected_indices and len(selected_indices) > 1:
                self.open_selection_menu(index)
                return
            path = self.filesystem_model.filePath(index)
            if os.path.isfile(path):
                self.open_file_menu(index)
                return
            if os.path.isdir(path):
                self.open_folder_menu(index)
                return

    def open_no_item_menu(self):
        """Open context menu when clicking on no item."""
        path = self.filesystem_model.rootPath()
        self.NoItemContextMenu.context_path = path
        self.NoItemContextMenu.popup(QtGui.QCursor.pos())

    def open_folder_menu(self, index):
        """Open context menu when clicking on a folder."""
        path = self.filesystem_model.filePath(index)
        self.FolderContextMenu.context_path = path
        self.FolderContextMenu.popup(QtGui.QCursor.pos())

    def open_file_menu(self, index):
        """Open context menu when clicking on a file."""
        path = self.filesystem_model.filePath(index)
        self.FileContextMenu.context_path = path
        self.FileContextMenu.popup(QtGui.QCursor.pos())

    def open_selection_menu(self, index):
        """Open context menu when clicking on a selection."""
        paths = self.selected_paths()
        self.SelectionContextMenu.context_path = paths
        self.SelectionContextMenu.popup(QtGui.QCursor.pos())

    def keyPressEvent(self, e):  # This is an overriding function # noqa: N802
        """Intercept keyPressEvent."""
        if e.key() == QtCore.Qt.Key_Delete and e.modifiers() == QtCore.Qt.NoModifier:
            self.delete(self.selected_paths())
        if e.key() == QtCore.Qt.Key_C and e.modifiers() == QtCore.Qt.ControlModifier:
            self.copy_to_clipboard(self.selected_paths())
        if e.key() == QtCore.Qt.Key_V and e.modifiers() == QtCore.Qt.ControlModifier:
            self.paste_from_clipboard(self.filesystem_model.rootPath())
        super().keyPressEvent(e)

    def selected_paths(self):
        """Return the paths of the currently selected items."""
        selected_paths = []
        for index in self.tree_view.selectionModel().selectedRows():
            if index.isValid():
                path = self.filesystem_model.filePath(index)
                selected_paths.append(path)
        return selected_paths

    # TODO: Possibly move the functions below to an own class or module as they dont directly relate to the FSExplorer
    def new_file_in_dir(self, head_path):
        """Prompts the user to ender a filename and creates that file in head_path."""
        tail_path, ok = QtWidgets.QInputDialog.getText(self, "Typstwriter", "Enter filename:", QtWidgets.QLineEdit.Normal)
        if ok and tail_path:
            path = os.path.join(head_path, tail_path)
            self.new_file(path)

    def new_file(self, path, overwrite=False):
        """Create new file in {path}."""
        if not os.path.exists(path) or overwrite:
            open(path, "w").close()
        else:
            logger.warning("File {!r} already exists. Will not overwrite.", path)
            QtWidgets.QMessageBox.warning(self, "Typstwriter", f"File '{path}' already exists.\nWill not overwrite.")

    def new_folder_in_dir(self, head_path):
        """Prompts the user to enter a foldername and creates that folder in head_path."""
        tail_path, ok = QtWidgets.QInputDialog.getText(self, "Typstwriter", "Enter foldername:", QtWidgets.QLineEdit.Normal)
        if ok and tail_path:
            path = os.path.join(head_path, tail_path)
            self.new_folder(path)

    def new_folder(self, path, overwrite=False):
        """Create new folder in {path}."""
        if not os.path.exists(path) or overwrite:
            os.makedirs(path)
        else:
            logger.warning("Folder {!r} already exists. Will not overwrite.", path)
            QtWidgets.QMessageBox.warning(self, "Typstwriter", f"Folder '{path}' already exists.\nWill not overwrite.")

    def rename_from(self, path_from):
        """Prompts the user to enter a new name."""
        head_path, tail_path_from = os.path.split(path_from)
        tail_path_to, ok = QtWidgets.QInputDialog.getText(
            self, "Typstwriter", "Enter new name:", QtWidgets.QLineEdit.Normal, text=tail_path_from
        )
        if ok and tail_path_to:
            path_to = os.path.join(head_path, tail_path_to)
            self.rename(path_from, path_to)

    def rename(self, path_from, path_to, overwrite=False):
        """Rename a file or folder from path_from to path_to."""
        if not os.path.exists(path_to) or overwrite:
            os.rename(path_from, path_to)
        else:
            logger.warning("{!r} already exists. Will not overwrite.", path_to)
            QtWidgets.QMessageBox.warning(self, "Typstwriter", f"'{path_to}' already exists.\nWill not overwrite.")

    def copy_to_clipboard(self, paths):
        """Copy a list of files or foler to the clipboard."""
        if not isinstance(paths, list):
            paths = [paths]

        urls = [QtCore.QUrl.fromLocalFile(path) for path in paths if os.path.exists(path)]

        if urls:
            mime_data = QtCore.QMimeData()
            mime_data.setUrls(urls)
            QtGui.QGuiApplication.clipboard().setMimeData(mime_data)

    def paste_from_clipboard(self, path):
        """Paste a file or folder from cliboard into path."""
        mime_data = QtGui.QGuiApplication.clipboard().mimeData()

        if mime_data.hasUrls():
            for uri in mime_data.data("text/uri-list").data().decode().split():
                fs = QtCore.QUrl(uri).toLocalFile()
                if os.path.exists(fs):
                    head, tail = os.path.split(fs)
                    path_to = os.path.join(path, tail)

                    if os.path.exists(path_to):
                        new_tail, ok = QtWidgets.QInputDialog.getText(
                            self,
                            "Typstwriter",
                            "Destination already exists. Enter new name:",
                            QtWidgets.QLineEdit.Normal,
                            text=tail,
                        )
                        path_to = os.path.join(path, new_tail)
                        if not ok:
                            return

                    self.copy_from_to(fs, path_to)

    def copy_from_to(self, path_from, path_to, overwrite=False):
        """Copy a file or folder from path_from to path_to."""
        if not os.path.isdir(os.path.split(path_to)[0]):
            return

        if not os.path.exists(path_from):
            return

        if not overwrite and os.path.exists(path_to):
            logger.warning("{!r} already exists. Will not overwrite.", path_to)
            QtWidgets.QMessageBox.warning(self, "Typstwriter", f"'{path_to}' already exists.\nWill not overwrite.")
            return

        if os.path.isdir(path_from):
            shutil.copytree(path_from, path_to, dirs_exist_ok=overwrite)
        else:
            shutil.copy2(path_from, path_to)

    def delete(self, paths):
        """Delete a folder or file."""
        if not isinstance(paths, list):
            paths = [paths]

        for path in paths:
            msg = QtWidgets.QMessageBox.question(self, "Typstwriter", f"Should '{path}' be deleted?")
            if msg == QtWidgets.QMessageBox.StandardButton.Yes:
                deletion_succeeded = QtCore.QFile.moveToTrash(path)
                if not deletion_succeeded:
                    logger.warning("Could not move {!r} to trash.", path)
